# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datapath.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x64\x61tapath.proto\x12\x14\x63om.deno.kv.datapath\"?\n\x0cSnapshotRead\x12/\n\x06ranges\x18\x01 \x03(\x0b\x32\x1f.com.deno.kv.datapath.ReadRange\"\xd9\x01\n\x12SnapshotReadOutput\x12\x35\n\x06ranges\x18\x01 \x03(\x0b\x32%.com.deno.kv.datapath.ReadRangeOutput\x12\x15\n\rread_disabled\x18\x02 \x01(\x08\x12#\n\x1bread_is_strongly_consistent\x18\x04 \x01(\x08\x12\x38\n\x06status\x18\x08 \x01(\x0e\x32(.com.deno.kv.datapath.SnapshotReadStatusJ\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08\"G\n\tReadRange\x12\r\n\x05start\x18\x01 \x01(\x0c\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x0c\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x0f\n\x07reverse\x18\x04 \x01(\x08\"@\n\x0fReadRangeOutput\x12-\n\x06values\x18\x01 \x03(\x0b\x32\x1d.com.deno.kv.datapath.KvEntry\"\x9e\x01\n\x0b\x41tomicWrite\x12+\n\x06\x63hecks\x18\x01 \x03(\x0b\x32\x1b.com.deno.kv.datapath.Check\x12\x31\n\tmutations\x18\x02 \x03(\x0b\x32\x1e.com.deno.kv.datapath.Mutation\x12/\n\x08\x65nqueues\x18\x03 \x03(\x0b\x32\x1d.com.deno.kv.datapath.Enqueue\"\x7f\n\x11\x41tomicWriteOutput\x12\x37\n\x06status\x18\x01 \x01(\x0e\x32\'.com.deno.kv.datapath.AtomicWriteStatus\x12\x14\n\x0cversionstamp\x18\x02 \x01(\x0c\x12\x15\n\rfailed_checks\x18\x04 \x03(\rJ\x04\x08\x03\x10\x04\"*\n\x05\x43heck\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\x14\n\x0cversionstamp\x18\x02 \x01(\x0c\"\xcb\x01\n\x08Mutation\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.deno.kv.datapath.KvValue\x12\x39\n\rmutation_type\x18\x03 \x01(\x0e\x32\".com.deno.kv.datapath.MutationType\x12\x14\n\x0c\x65xpire_at_ms\x18\x04 \x01(\x03\x12\x0f\n\x07sum_min\x18\x05 \x01(\x0c\x12\x0f\n\x07sum_max\x18\x06 \x01(\x0c\x12\x11\n\tsum_clamp\x18\x07 \x01(\x08\"N\n\x07KvValue\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x35\n\x08\x65ncoding\x18\x02 \x01(\x0e\x32#.com.deno.kv.datapath.ValueEncoding\"r\n\x07KvEntry\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x35\n\x08\x65ncoding\x18\x03 \x01(\x0e\x32#.com.deno.kv.datapath.ValueEncoding\x12\x14\n\x0cversionstamp\x18\x04 \x01(\x0c\"f\n\x07\x45nqueue\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x13\n\x0b\x64\x65\x61\x64line_ms\x18\x02 \x01(\x03\x12\x1b\n\x13keys_if_undelivered\x18\x03 \x03(\x0c\x12\x18\n\x10\x62\x61\x63koff_schedule\x18\x04 \x03(\r\"5\n\x05Watch\x12,\n\x04keys\x18\x01 \x03(\x0b\x32\x1e.com.deno.kv.datapath.WatchKey\"{\n\x0bWatchOutput\x12\x38\n\x06status\x18\x01 \x01(\x0e\x32(.com.deno.kv.datapath.SnapshotReadStatus\x12\x32\n\x04keys\x18\x02 \x03(\x0b\x32$.com.deno.kv.datapath.WatchKeyOutput\"\x17\n\x08WatchKey\x12\x0b\n\x03key\x18\x01 \x01(\x0c\"Z\n\x0eWatchKeyOutput\x12\x0f\n\x07\x63hanged\x18\x01 \x01(\x08\x12\x37\n\x10\x65ntry_if_changed\x18\x02 \x01(\x0b\x32\x1d.com.deno.kv.datapath.KvEntry*N\n\x12SnapshotReadStatus\x12\x12\n\x0eSR_UNSPECIFIED\x10\x00\x12\x0e\n\nSR_SUCCESS\x10\x01\x12\x14\n\x10SR_READ_DISABLED\x10\x02*\x80\x01\n\x0cMutationType\x12\x11\n\rM_UNSPECIFIED\x10\x00\x12\t\n\x05M_SET\x10\x01\x12\x0c\n\x08M_DELETE\x10\x02\x12\t\n\x05M_SUM\x10\x03\x12\t\n\x05M_MAX\x10\x04\x12\t\n\x05M_MIN\x10\x05\x12#\n\x1fM_SET_SUFFIX_VERSIONSTAMPED_KEY\x10\t*I\n\rValueEncoding\x12\x12\n\x0eVE_UNSPECIFIED\x10\x00\x12\t\n\x05VE_V8\x10\x01\x12\x0b\n\x07VE_LE64\x10\x02\x12\x0c\n\x08VE_BYTES\x10\x03*v\n\x11\x41tomicWriteStatus\x12\x12\n\x0e\x41W_UNSPECIFIED\x10\x00\x12\x0e\n\nAW_SUCCESS\x10\x01\x12\x14\n\x10\x41W_CHECK_FAILURE\x10\x02\x12\x15\n\x11\x41W_WRITE_DISABLED\x10\x05\"\x04\x08\x03\x10\x03\"\x04\x08\x04\x10\x04\"\x04\x08\x06\x10\x06\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'datapath_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_SNAPSHOTREADSTATUS']._serialized_start=1601
  _globals['_SNAPSHOTREADSTATUS']._serialized_end=1679
  _globals['_MUTATIONTYPE']._serialized_start=1682
  _globals['_MUTATIONTYPE']._serialized_end=1810
  _globals['_VALUEENCODING']._serialized_start=1812
  _globals['_VALUEENCODING']._serialized_end=1885
  _globals['_ATOMICWRITESTATUS']._serialized_start=1887
  _globals['_ATOMICWRITESTATUS']._serialized_end=2005
  _globals['_SNAPSHOTREAD']._serialized_start=40
  _globals['_SNAPSHOTREAD']._serialized_end=103
  _globals['_SNAPSHOTREADOUTPUT']._serialized_start=106
  _globals['_SNAPSHOTREADOUTPUT']._serialized_end=323
  _globals['_READRANGE']._serialized_start=325
  _globals['_READRANGE']._serialized_end=396
  _globals['_READRANGEOUTPUT']._serialized_start=398
  _globals['_READRANGEOUTPUT']._serialized_end=462
  _globals['_ATOMICWRITE']._serialized_start=465
  _globals['_ATOMICWRITE']._serialized_end=623
  _globals['_ATOMICWRITEOUTPUT']._serialized_start=625
  _globals['_ATOMICWRITEOUTPUT']._serialized_end=752
  _globals['_CHECK']._serialized_start=754
  _globals['_CHECK']._serialized_end=796
  _globals['_MUTATION']._serialized_start=799
  _globals['_MUTATION']._serialized_end=1002
  _globals['_KVVALUE']._serialized_start=1004
  _globals['_KVVALUE']._serialized_end=1082
  _globals['_KVENTRY']._serialized_start=1084
  _globals['_KVENTRY']._serialized_end=1198
  _globals['_ENQUEUE']._serialized_start=1200
  _globals['_ENQUEUE']._serialized_end=1302
  _globals['_WATCH']._serialized_start=1304
  _globals['_WATCH']._serialized_end=1357
  _globals['_WATCHOUTPUT']._serialized_start=1359
  _globals['_WATCHOUTPUT']._serialized_end=1482
  _globals['_WATCHKEY']._serialized_start=1484
  _globals['_WATCHKEY']._serialized_end=1507
  _globals['_WATCHKEYOUTPUT']._serialized_start=1509
  _globals['_WATCHKEYOUTPUT']._serialized_end=1599
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import time
from datetime import datetime


def loop_time(wall_time: datetime | float | None = None) -> float:
    """
    Get time in the running event loop's arbitrary time offset.

    If `wall_time` is set, the returned time is the absolute loop time that this
    wall-clock time will occur at. If `wall_time` is not set, the current loop
    time is returned.

    Notes
    -----
    Loop time does not correspond to wall-clock time because it uses
    time.monotonic, which has an arbitrary epoch/offset.
    """
    if wall_time is None:
        return asyncio.get_running_loop().time()
    if isinstance(wall_time, datetime):
        wall_time = wall_time.timestamp()
    return wall_time - loop_time_offset()


def loop_time_offset() -> float:
    """
    Get the number of seconds to be added to loop_time() to get time.time().

    loop_time() + loop_time_offset() should equal time.time()
    """
    return time.time() - loop_time()



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SnapshotReadStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    SR_UNSPECIFIED: _ClassVar[SnapshotReadStatus]
    SR_SUCCESS: _ClassVar[SnapshotReadStatus]
    SR_READ_DISABLED: _ClassVar[SnapshotReadStatus]

class MutationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    M_UNSPECIFIED: _ClassVar[MutationType]
    M_SET: _ClassVar[MutationType]
    M_DELETE: _ClassVar[MutationType]
    M_SUM: _ClassVar[MutationType]
    M_MAX: _ClassVar[MutationType]
    M_MIN: _ClassVar[MutationType]
    M_SET_SUFFIX_VERSIONSTAMPED_KEY: _ClassVar[MutationType]

class ValueEncoding(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    VE_UNSPECIFIED: _ClassVar[ValueEncoding]
    VE_V8: _ClassVar[ValueEncoding]
    VE_LE64: _ClassVar[ValueEncoding]
    VE_BYTES: _ClassVar[ValueEncoding]

class AtomicWriteStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    AW_UNSPECIFIED: _ClassVar[AtomicWriteStatus]
    AW_SUCCESS: _ClassVar[AtomicWriteStatus]
    AW_CHECK_FAILURE: _ClassVar[AtomicWriteStatus]
    AW_WRITE_DISABLED: _ClassVar[AtomicWriteStatus]
SR_UNSPECIFIED: SnapshotReadStatus
SR_SUCCESS: SnapshotReadStatus
SR_READ_DISABLED: SnapshotReadStatus
M_UNSPECIFIED: MutationType
M_SET: MutationType
M_DELETE: MutationType
M_SUM: MutationType
M_MAX: MutationType
M_MIN: MutationType
M_SET_SUFFIX_VERSIONSTAMPED_KEY: MutationType
VE_UNSPECIFIED: ValueEncoding
VE_V8: ValueEncoding
VE_LE64: ValueEncoding
VE_BYTES: ValueEncoding
AW_UNSPECIFIED: AtomicWriteStatus
AW_SUCCESS: AtomicWriteStatus
AW_CHECK_FAILURE: AtomicWriteStatus
AW_WRITE_DISABLED: AtomicWriteStatus

class SnapshotRead(_message.Message):
    __slots__ = ["ranges"]
    RANGES_FIELD_NUMBER: _ClassVar[int]
    ranges: _containers.RepeatedCompositeFieldContainer[ReadRange]
    def __init__(self, ranges: _Optional[_Iterable[_Union[ReadRange, _Mapping]]] = ...) -> None: ...

class SnapshotReadOutput(_message.Message):
    __slots__ = ["ranges", "read_disabled", "read_is_strongly_consistent", "status"]
    RANGES_FIELD_NUMBER: _ClassVar[int]
    READ_DISABLED_FIELD_NUMBER: _ClassVar[int]
    READ_IS_STRONGLY_CONSISTENT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ranges: _containers.RepeatedCompositeFieldContainer[ReadRangeOutput]
    read_disabled: bool
    read_is_strongly_consistent: bool
    status: SnapshotReadStatus
    def __init__(self, ranges: _Optional[_Iterable[_Union[ReadRangeOutput, _Mapping]]] = ..., read_disabled: bool = ..., read_is_strongly_consistent: bool = ..., status: _Optional[_Union[SnapshotReadStatus, str]] = ...) -> None: ...

class ReadRange(_message.Message):
    __slots__ = ["start", "end", "limit", "reverse"]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    REVERSE_FIELD_NUMBER: _ClassVar[int]
    start: bytes
    end: bytes
    limit: int
    reverse: bool
    def __init__(self, start: _Optional[bytes] = ..., end: _Optional[bytes] = ..., limit: _Optional[int] = ..., reverse: bool = ...) -> None: ...

class ReadRangeOutput(_message.Message):
    __slots__ = ["values"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[KvEntry]
    def __init__(self, values: _Optional[_Iterable[_Union[KvEntry, _Mapping]]] = ...) -> None: ...

class AtomicWrite(_message.Message):
    __slots__ = ["checks", "mutations", "enqueues"]
    CHECKS_FIELD_NUMBER: _ClassVar[int]
    MUTATIONS_FIELD_NUMBER: _ClassVar[int]
    ENQUEUES_FIELD_NUMBER: _ClassVar[int]
    checks: _containers.RepeatedCompositeFieldContainer[Check]
    mutations: _containers.RepeatedCompositeFieldContainer[Mutation]
    enqueues: _containers.RepeatedCompositeFieldContainer[Enqueue]
    def __init__(self, checks: _Optional[_Iterable[_Union[Check, _Mapping]]] = ..., mutations: _Optional[_Iterable[_Union[Mutation, _Mapping]]] = ..., enqueues: _Optional[_Iterable[_Union[Enqueue, _Mapping]]] = ...) -> None: ...

class AtomicWriteOutput(_message.Message):
    __slots__ = ["status", "versionstamp", "failed_checks"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    VERSIONSTAMP_FIELD_NUMBER: _ClassVar[int]
    FAILED_CHECKS_FIELD_NUMBER: _ClassVar[int]
    status: AtomicWriteStatus
    versionstamp: bytes
    failed_checks: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, status: _Optional[_Union[AtomicWriteStatus, str]] = ..., versionstamp: _Optional[bytes] = ..., failed_checks: _Optional[_Iterable[int]] = ...) -> None: ...

class Check(_message.Message):
    __slots__ = ["key", "versionstamp"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VERSIONSTAMP_FIELD_NUMBER: _ClassVar[int]
    key: bytes
    versionstamp: bytes
    def __init__(self, key: _Optional[bytes] = ..., versionstamp: _Optional[bytes] = ...) -> None: ...

class Mutation(_message.Message):
    __slots__ = ["key", "value", "mutation_type", "expire_at_ms", "sum_min", "sum_max", "sum_clamp"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    MUTATION_TYPE_FIELD_NUMBER: _ClassVar[int]
    EXPIRE_AT_MS_FIELD_NUMBER: _ClassVar[int]
    SUM_MIN_FIELD_NUMBER: _ClassVar[int]
    SUM_MAX_FIELD_NUMBER: _ClassVar[int]
    SUM_CLAMP_FIELD_NUMBER: _ClassVar[int]
    key: bytes
    value: KvValue
    mutation_type: MutationType
    expire_at_ms: int
    sum_min: bytes
    sum_max: bytes
    sum_clamp: bool
    def __init__(self, key: _Optional[bytes] = ..., value: _Optional[_Union[KvValue, _Mapping]] = ..., mutation_type: _Optional[_Union[MutationType, str]] = ..., expire_at_ms: _Optional[int] = ..., sum_min: _Optional[bytes] = ..., sum_max: _Optional[bytes] = ..., sum_clamp: bool = ...) -> None: ...

class KvValue(_message.Message):
    __slots__ = ["data", "encoding"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    ENCODING_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    encoding: ValueEncoding
    def __init__(self, data: _Optional[bytes] = ..., encoding: _Optional[_Union[ValueEncoding, str]] = ...) -> None: ...

class KvEntry(_message.Message):
    __slots__ = ["key", "value", "encoding", "versionstamp"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    ENCODING_FIELD_NUMBER: _ClassVar[int]
    VERSIONSTAMP_FIELD_NUMBER: _ClassVar[int]
    key: bytes
    value: bytes
    encoding: ValueEncoding
    versionstamp: bytes
    def __init__(self, key: _Optional[bytes] = ..., value: _Optional[bytes] = ..., encoding: _Optional[_Union[ValueEncoding, str]] = ..., versionstamp: _Optional[bytes] = ...) -> None: ...

class Enqueue(_message.Message):
    __slots__ = ["payload", "deadline_ms", "keys_if_undelivered", "backoff_schedule"]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    DEADLINE_MS_FIELD_NUMBER: _ClassVar[int]
    KEYS_IF_UNDELIVERED_FIELD_NUMBER: _ClassVar[int]
    BACKOFF_SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    payload: bytes
    deadline_ms: int
    keys_if_undelivered: _containers.RepeatedScalarFieldContainer[bytes]
    backoff_schedule: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, payload: _Optional[bytes] = ..., deadline_ms: _Optional[int] = ..., keys_if_undelivered: _Optional[_Iterable[bytes]] = ..., backoff_schedule: _Optional[_Iterable[int]] = ...) -> None: ...

class Watch(_message.Message):
    __slots__ = ["keys"]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    keys: _containers.RepeatedCompositeFieldContainer[WatchKey]
    def __init__(self, keys: _Optional[_Iterable[_Union[WatchKey, _Mapping]]] = ...) -> None: ...

class WatchOutput(_message.Message):
    __slots__ = ["status", "keys"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    status: SnapshotReadStatus
    keys: _containers.RepeatedCompositeFieldContainer[WatchKeyOutput]
    def __init__(self, status: _Optional[_Union[SnapshotReadStatus, str]] = ..., keys: _Optional[_Iterable[_Union[WatchKeyOutput, _Mapping]]] = ...) -> None: ...

class WatchKey(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: bytes
    def __init__(self, key: _Optional[bytes] = ...) -> None: ...

class WatchKeyOutput(_message.Message):
    __slots__ = ["changed", "entry_if_changed"]
    CHANGED_FIELD_NUMBER: _ClassVar[int]
    ENTRY_IF_CHANGED_FIELD_NUMBER: _ClassVar[int]
    changed: bool
    entry_if_changed: KvEntry
    def __init__(self, changed: bool = ..., entry_if_changed: _Optional[_Union[KvEntry, _Mapping]] = ...) -> None: ...



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Mapping
from typing import Sequence
from typing import cast
from uuid import UUID

import aiohttp
import aiohttp.client_exceptions
from yarl import URL

from denokv._pycompat.dataclasses import slots_if310
from denokv._pycompat.enum import StrEnum
from denokv._rfc3339 import parse_rfc3339_datetime
from denokv.errors import DenoKvError
from denokv.errors import DenoKvValidationError
from denokv.result import Err
from denokv.result import Ok
from denokv.result import Result


@dataclass(frozen=True, **slots_if310())
class DatabaseMetadata:
    """
    MetadataExchangeResponse.

    ```json
    {
      "version": 2,
      "uuid": "a1b2c3d4-e5f6-7g8h-9i1j-2k3l4m5n6o7p",
      "endpoints": [
        {
          "url": "/v2",
          "consistency": "strong"
        },
        {
          "url": "https://mirror.example.com/v2",
          "consistency": "eventual"
        }
      ],
      "token": "123abc456def789ghi",
      "expiresAt": "2023-10-01T00:00:00Z"
    }
    ```
    """

    version: int
    database_id: UUID
    """Note: actual responses use "uuid"."""
    endpoints: Sequence[EndpointInfo]
    token: str
    expires_at: datetime


class ConsistencyLevel(StrEnum):
    STRONG = "strong"
    EVENTUAL = "eventual"


@dataclass(frozen=True, **slots_if310())
class EndpointInfo:
    url: URL
    consistency: ConsistencyLevel


@dataclass(init=False)
class InvalidMetadataResponseDenoKvError(DenoKvValidationError):
    data: object

    def __init__(self, message: str, data: object, *args: object) -> None:
        super().__init__(message, data, *args)

    @property  # type: ignore[no-redef]
    def data(self) -> object:
        return self.args[1]


def read_metadata_exchange_response(
    data: object, *, base_url: URL
) -> Result[DatabaseMetadata, InvalidMetadataResponseDenoKvError]:
    # This is caused by using the fn incorrectly, not a data error, so throw)
    if not base_url.is_absolute():
        raise TypeError("base_url is not absolute")

    Error = InvalidMetadataResponseDenoKvError
    if not isinstance(data, Mapping):
        return Err(Error("JSON value is not an object", data))
    version = data.get("version")
    if not (isinstance(version, int) and 1 <= version <= 3):
        return Err(Error(f"unsupported version: {version!r}", data))

    raw_database_id = data.get("databaseId") or data.get("uuid")
    try:
        database_id = UUID(raw_database_id)
    except ValueError as e:
        err = Error(f"databaseId/uuid is not a UUID: {raw_database_id!r}", data=data)
        err.__cause__ = e
        return Err(err)

    token = data.get("token")
    if not (isinstance(token, str) and len(token) > 0):
        return Err(Error("token is not a non-empty string", data))

    raw_expires_at = data.get("expiresAt")
    parsed_expires_at = None
    if isinstance(raw_expires_at, str):
        parsed_expires_at = parse_rfc3339_datetime(raw_expires_at)

    if not isinstance(parsed_expires_at, Ok):
        err = Error(
            f"expiresAt is not an RFC3339 date-time: {raw_expires_at!r}", data=data
        )
        if isinstance(parsed_expires_at, Err):
            err.__cause__ = parsed_expires_at.error
        return Err(err)

    raw_endpoints = data.get("endpoints")
    if not isinstance(raw_endpoints, Sequence):
        return Err(Error("endpoints is not an array", data))
    endpoints: list[EndpointInfo] = []
    for i, raw_endpoint in enumerate(raw_endpoints):
        if not isinstance(raw_endpoint, Mapping):
            return Err(Error(f"endpoints[{i}] is not an object", data))
        raw_url = raw_endpoint.get("url")
        try:
            if not isinstance(raw_url, str):
                raise TypeError("url is not a string")
            url = base_url.join(URL(raw_url))
            if url.scheme not in ("http", "https"):
                raise ValueError("scheme must be http or https")
            # TODO: JSON schema comments that URL must not end with /. Should we
            #   validate/enforce this? or later when using URLs?
        except (TypeError, ValueError) as e:
            err = Error(f"endpoints[{i}].url is invalid", data=data)
            err.__cause__ = e
            return Err(err)

        raw_consistency = raw_endpoint.get("consistency")
        try:
            if not isinstance(raw_consistency, str):
                raise TypeError("value must be a string")
            consistency = ConsistencyLevel(raw_consistency)
        except (TypeError, ValueError):
            return Err(
                Error(
                    f"endpoints[{i}].consistency is not one of "
                    f"{', '.join(ConsistencyLevel)}",
                    data=data,
                )
            )

        endpoints.append(EndpointInfo(url=url, consistency=consistency))

    return Ok(
        DatabaseMetadata(
            version=version,
            database_id=database_id,
            endpoints=tuple(endpoints),
            token=token,
            expires_at=parsed_expires_at.value,
        )
    )


@dataclass(init=False)
class MetadataExchangeDenoKvError(DenoKvError):
    retryable: bool

    def __init__(self, message: str, *args: object, retryable: bool) -> None:
        self.retryable = retryable
        super().__init__(message, *args)


@dataclass(init=False)
class InvalidMetadataExchangeDenoKvError(MetadataExchangeDenoKvError):
    """Metadata received from the KV server is not valid."""

    data: object | DatabaseMetadata
    """The invalid JSON data, or a parsed-but-invalid data."""

    def __init__(
        self,
        message: str,
        *args: object,
        data: object | DatabaseMetadata,
        retryable: bool,
    ) -> None:
        self.data = data
        super().__init__(message, *args, retryable=retryable)


@dataclass(init=False)
class HttpResponseMetadataExchangeDenoKvError(MetadataExchangeDenoKvError):
    """The KV server responded to the metadata exchange request unsuccessfully."""

    status: int
    body_text: str

    def __init__(
        self, message: str, status: int, body_text: str, *args: object, retryable: bool
    ) -> None:
        super().__init__(message, status, body_text, *args, retryable=retryable)

    @property  # type: ignore[no-redef]
    def status(self) -> int:
        return cast(int, self.args[1])

    @property  # type: ignore[no-redef]
    def body_text(self) -> str:
        return cast(str, self.args[2])


class HttpRequestMetadataExchangeDenoKvError(MetadataExchangeDenoKvError):
    """Unable to make an HTTP request to exchange metadata with the KV server."""

    pass


async def get_database_metadata(
    *,
    session: aiohttp.ClientSession,
    server_url: str | URL,
    access_token: str,
) -> Result[DatabaseMetadata, MetadataExchangeDenoKvError]:
    err: MetadataExchangeDenoKvError
    try:
        async with session.post(
            url=server_url,
            json={"supportedVersions": [1, 2, 3]},
            allow_redirects=True,
            headers={"Authorization": f"Bearer {access_token}"},
        ) as response:
            if not response.ok:
                body_text = await response.text()
                if 400 <= response.status < 500:
                    return Err(
                        HttpResponseMetadataExchangeDenoKvError(
                            "Server rejected metadata exchange request indicating "
                            "client error",
                            status=response.status,
                            body_text=body_text,
                            retryable=False,
                        )
                    )
                elif 500 <= response.status < 600:
                    return Err(
                        HttpResponseMetadataExchangeDenoKvError(
                            "Server failed to respond to metadata exchange request "
                            "indicating server error",
                            status=response.status,
                            body_text=body_text,
                            retryable=True,
                        )
                    )
                else:
                    return Err(
                        HttpResponseMetadataExchangeDenoKvError(
                            "Server responded to metadata exchange request with "
                            "unexpected status",
                            status=response.status,
                            body_text=body_text,
                            retryable=False,
                        )
                    )
            data = await response.json()
            base_url = response.history[-1].url if response.history else URL(server_url)
    except aiohttp.client_exceptions.ClientError as e:
        # We don't retry invalid URLs. Invalid URL could be the server_url arg,
        # or we could have been redirected to an invalid URL. The kv-connect
        # spec requires that we abort if redirected to an invalid URL.
        retryable = not isinstance(e, aiohttp.InvalidURL)
        err = HttpRequestMetadataExchangeDenoKvError(
            "Failed to make HTTP request to KV server to exchange metadata",
            retryable=retryable,
        )
        err.__cause__ = e
        return Err(err)

    result = read_metadata_exchange_response(data, base_url=base_url)
    if isinstance(result, Err):
        err = InvalidMetadataExchangeDenoKvError(
            "Server responded to metadata exchange with invalid metadata",
            data=data,
            retryable=False,
        )
        err.__cause__ = result.error
        return Err(err)
    meta = result.value

    # Semantic validation
    if not any(e.consistency == ConsistencyLevel.STRONG for e in meta.endpoints):
        return Err(
            InvalidMetadataExchangeDenoKvError(
                f"Server responded to metadata exchange without any "
                f"{ConsistencyLevel.STRONG} consistency endpoints",
                data=meta,
                retryable=False,
            )
        )

    return Ok(meta)



from dataclasses import dataclass
from typing import cast


@dataclass(init=False)
class DenoKvError(BaseException):
    message: str

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)
        if not isinstance(message, str):
            raise TypeError(f"first argument must be a str message: {message!r}")

    @property  # type: ignore[no-redef]
    def message(self) -> str:
        return cast(str, self.args[0])


class DenoKvValidationError(ValueError, DenoKvError):
    pass


class DenoKvUserError(DenoKvError):
    """An error caused by bad user input."""


@dataclass(init=False)
class InvalidCursor(DenoKvUserError):
    """A cursor string passed to [kv.list()] is invalid."""

    cursor: str

    def __init__(self, message: str, *args: object, cursor: str) -> None:
        super().__init__(message, *args)
        self.cursor = cursor



from __future__ import annotations

from denokv.auth import ConsistencyLevel as ConsistencyLevel
from denokv.auth import MetadataExchangeDenoKvError as MetadataExchangeDenoKvError
from denokv.datapath import AnyKvKey as AnyKvKey
from denokv.datapath import DataPathDenoKvError as DataPathDenoKvError
from denokv.datapath import KvKeyEncodable as KvKeyEncodable
from denokv.datapath import KvKeyPiece as KvKeyPiece
from denokv.datapath import KvKeyTuple as KvKeyTuple
from denokv.errors import DenoKvError as DenoKvError
from denokv.errors import DenoKvUserError as DenoKvUserError
from denokv.errors import InvalidCursor as InvalidCursor
from denokv.kv import CursorFormatType as CursorFormatType
from denokv.kv import Kv as Kv
from denokv.kv import KvCredentials as KvCredentials
from denokv.kv import KvEntry as KvEntry
from denokv.kv import KvListOptions as KvListOptions
from denokv.kv import KvU64 as KvU64
from denokv.kv import ListKvEntry as ListKvEntry
from denokv.kv import VersionStamp as VersionStamp
from denokv.kv import open_kv as open_kv
from denokv.kv_keys import KvKey as KvKey



from __future__ import annotations

import functools
import sys
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any
from typing import Final
from typing import Generic
from typing import Iterator
from typing import Sequence
from typing import Union
from typing import cast
from typing import overload

from fdb.tuple import pack
from fdb.tuple import unpack

from denokv._pycompat.dataclasses import slots_if310
from denokv._pycompat.typing import override
from denokv.datapath import KV_KEY_PIECE_TYPES
from denokv.datapath import AnyKvKey
from denokv.datapath import AnyKvKeyT
from denokv.datapath import AnyKvKeyT_co
from denokv.datapath import KvKeyEncodable
from denokv.datapath import KvKeyEncodableT
from denokv.datapath import KvKeyPiece
from denokv.datapath import KvKeyRangeEncodable
from denokv.datapath import KvKeyTuple
from denokv.datapath import PackKeyRangeOptions
from denokv.datapath import is_any_kv_key
from denokv.datapath import is_kv_key_tuple
from denokv.datapath import pack_key
from denokv.datapath import pack_key_range
from denokv.result import Nothing
from denokv.result import Some

if TYPE_CHECKING:
    from typing_extensions import Self
    from typing_extensions import SupportsIndex
    from typing_extensions import TypeAlias
    from typing_extensions import TypeVar
    from typing_extensions import TypeVarTuple
    from typing_extensions import Unpack

    T = TypeVar("T", default=object)
    # Note that the default arg doesn't seem to work with MyPy yet. The
    # DefaultKvKey alias is what this should behave as when defaulted.
    Pieces = TypeVarTuple("Pieces", default=Unpack[tuple[KvKeyPiece, ...]])
else:
    from typing import TypeVar

    T = TypeVar("T")
    Unpack = tuple  # hack to support py39 at runtime w/o typing_extensions
    Pieces = TypeVar("Pieces")  # hack to support py39 at runtime w/o typing_extensions

_T_co = TypeVar("_T_co", covariant=True)


class _KvKeyState:
    _unpacked: KvKeyTuple
    _packed: bytes | None
    __slots__ = ("_unpacked", "_packed")

    def __init__(self, *pieces: KvKeyPiece) -> None:
        self._unpacked = pieces
        if not is_kv_key_tuple(pieces):
            raise TypeError(
                f"key contains types other than "
                f"{', '.join(t.__name__ for t in KV_KEY_PIECE_TYPES)}: {pieces!r}"
            )
        self._packed = None

    def __eq__(self, value: object) -> bool:
        # We don't compare equal to plain key tuples, because the comparison
        # is not symmetric and tuple's hash does not follow this equality.
        if not isinstance(value, KvKeyEncodable):
            return NotImplemented
        # unpacked tuples don't compare correctly because python == and hash
        # treats int and integer floats as the same, but FDB keys don't.
        return self.kv_key_bytes() == pack_key(value)

    def __hash__(self) -> int:
        return hash((KvKeyEncodable, self.kv_key_bytes()))

    def __repr__(self) -> str:
        return f"KvKey{self._unpacked!r}"

    @overload
    @classmethod
    def wrap_tuple_keys(cls, key: KvKeyTuple) -> DefaultKvKey: ...

    @overload
    @classmethod
    def wrap_tuple_keys(cls, key: KvKeyEncodableT) -> KvKeyEncodableT: ...

    @classmethod
    def wrap_tuple_keys(
        cls, key: KvKeyTuple | KvKeyEncodableT
    ) -> DefaultKvKey | KvKeyEncodableT:
        """
        Return simple tuples as KvKey but return [KvKeyEncodable] keys as-is.

        Notes
        -----
        By wrapping plain tuple keys as KvKey we preserve int as int if a key
        read from the DB is passed back into a Kv method which is normalising int
        to float for JS compatibility.
        """
        if isinstance(key, KvKeyEncodable):
            return key  # type: ignore[return-value,unused-ignore]
        return cls(*key)  # type: ignore[arg-type,return-value,unused-ignore]

    def kv_key_bytes(self) -> bytes:
        if (packed := self._packed) is None:
            self._packed = packed = pack(self._unpacked)
        return packed

    def __bytes__(self) -> bytes:
        return self.kv_key_bytes()

    @classmethod
    def from_kv_key_bytes(cls, packed_key: bytes) -> DefaultKvKey:
        """Create a KvKey by unpacking a packed key."""
        try:
            # If packed key contains types other than allowed by KvKeyPiece
            # then the constructor throws TypeError, so this is type-safe.
            kvkey = cls(*unpack(packed_key))  # type: ignore[arg-type]
        except ValueError as e:
            raise ValueError(
                f"Cannot create {cls.__name__} from packed key: {packed_key!r}:"
                f" value is not a valid packed key"
            ) from e
        except TypeError as e:
            raise ValueError(
                f"Cannot create {cls.__name__} from packed key: {packed_key!r}: {e}"
            ) from e
        kvkey._packed = packed_key  # pre-cache the packed representation
        return kvkey  # type: ignore[return-value]

    def range_start(self, start: StartT) -> KvKeyRange[StartT, Exclude[Self]]:
        """Get a key range from a `start` boundary up to but excluding this key."""
        return KvKeyRange(start, Exclude(self))

    def range_stop(self, stop: StopT) -> KvKeyRange[Include[Self], StopT]:
        """Get a key range starting with this key up to the `stop` boundary."""
        return KvKeyRange(Include(self), stop)

    def range(self) -> KvKeyRange[Include[Self], IncludePrefix[Self]]:
        """Get a key range starting with this key, including all suffixes of it."""
        return KvKeyRange(Include(self), IncludePrefix(self))

    def include(self) -> Include[Self]:
        """Get this key as an inclusive key range boundary."""
        return Include(self)

    def include_prefix(self) -> IncludePrefix[Self]:
        """Get this key as an inclusive prefix key range boundary."""
        return IncludePrefix(self)

    def exclude(self) -> Exclude[Self]:
        """Get this key as an exclusive key range boundary."""
        return Exclude(self)

    # Methods supported by tuple that we handle slightly differently
    def __lt__(self, value: AnyKvKey, /) -> bool:
        try:
            return self.kv_key_bytes() < pack_key(value)
        except TypeError:
            return NotImplemented

    def __le__(self, value: AnyKvKey, /) -> bool:
        try:
            return self.kv_key_bytes() <= pack_key(value)
        except TypeError:
            return NotImplemented

    def __gt__(self, value: AnyKvKey, /) -> bool:
        try:
            return self.kv_key_bytes() > pack_key(value)
        except TypeError:
            return NotImplemented

    def __ge__(self, value: AnyKvKey, /) -> bool:
        try:
            return self.kv_key_bytes() >= pack_key(value)
        except TypeError:
            return NotImplemented


if TYPE_CHECKING:
    # KvKey isn't actually a tuple subclass at runtime, because we need to store
    # extra state on each instance to cache the packed representation. The tuple
    # type does not allow adding extra __slots__. However we do want to use
    # tuple-like per-element typing, but the only way to do this is to make the
    # type checker think this is a tuple subtype — Python typing has no way to
    # type a __getitem__ call as returning the per-item types — tuple is
    # special-cased.

    # We need ignore[misc] as _KvKeyState order methods like __le__ are broader.
    class KvKey(_KvKeyState, tuple[Unpack[Pieces]]):  # type: ignore[misc]
        """
        A key identifying a value in a Deno KV database.

        KvKey is a tuple of key pieces — str, bytes, int, float or bool. Unlike a
        plain tuple, KvKey's values are guaranteed to only be valid key values, and
        int values are not coerced to float for JavaScript comparability when used
        with [Kv] methods.

        [Kv]: `denokv.kv.Kv`
        """

        # The Pieces TypeVarTuple cannot be bounded to KvKeyPiece elements, so this
        # type can hold any element, but  only KvKeyPiece can exist at runtime.
        def __new__(cls, *pieces: Unpack[Pieces]) -> KvKey[Unpack[Pieces]]: ...

        @overload
        def __getitem__(self, index: SupportsIndex, /) -> KvKeyPiece: ...
        @overload
        def __getitem__(self, index: slice, /) -> KvKey[KvKeyTuple]: ...
        def __getitem__(
            self, index: slice | SupportsIndex, /
        ) -> KvKey | KvKeyPiece: ...

        def __add__(self, value: KvKeyTuple, /) -> KvKey: ...  # type: ignore[override]
        def __mul__(self, value: SupportsIndex, /) -> KvKey: ...
        def __rmul__(self, value: SupportsIndex, /) -> KvKey: ...

else:

    @Sequence.register
    class KvKey(_KvKeyState):
        """
        A key identifying a value in a Deno KV database.

        KvKey is a tuple of key pieces — str, bytes, int, float or bool. Unlike a
        plain tuple, KvKey's values are guaranteed to only be valid key values, and
        int values are not coerced to float for JavaScript comparability when used
        with [Kv] methods.

        [Kv]: `denokv.kv.Kv`
        """

        __slots__ = ("__weakref__",)

        def __len__(self) -> int:
            return len(self._unpacked)

        def __contains__(self, value: object, /) -> bool:
            return value in self._unpacked

        @overload
        def __getitem__(self, index: SupportsIndex, /) -> KvKeyPiece: ...

        @overload
        def __getitem__(self, index: slice, /) -> Self: ...

        def __getitem__(self, index: slice | SupportsIndex, /) -> Self | KvKeyPiece:
            if isinstance(index, slice):
                return KvKey(*self._unpacked[index])
            return self._unpacked[index]

        def __iter__(self) -> Iterator[KvKeyPiece]:
            return iter(self._unpacked)

        def __reversed__(self) -> Iterator[_T_co]:
            return reversed(self._unpacked)

        def __add__(self, value: AnyKvKey, /) -> KvKey[KvKeyTuple]:
            if isinstance(value, KvKey):
                return KvKey(*self._unpacked, *value._unpacked)
            if is_kv_key_tuple(value):
                return KvKey(*self._unpacked, *value)
            return KvKey(*self._unpacked, *unpack(pack_key(value)))

        def __mul__(self, value: SupportsIndex, /) -> KvKey[KvKeyTuple]:
            return KvKey(*(self._unpacked * value))

        def __rmul__(self, value: SupportsIndex, /) -> KvKey[KvKeyTuple]:
            return KvKey(*(self._unpacked * value))

        def count(self, value: Any, /) -> int:
            return self._unpacked.count(value)

        def index(
            self,
            value: object,
            start: int = 0,
            stop: int = sys.maxsize,
            /,
        ) -> int:
            return self._unpacked.index(value, start, stop)


# Ideally the default parameter of the Pieces KvKeyTuple would make this the
# default for KvKey (with no generic type), but mypy thinks it is
# KvKey[*tuple[Any, ...]] when used without generic type args.
DefaultKvKey: TypeAlias = "KvKey[Unpack[tuple[KvKeyPiece, ...]]]"
"""KvKey containing any number of key values of any allowed type."""


@dataclass(frozen=True, **slots_if310())
class _KeyBoundary(Generic[AnyKvKeyT_co]):
    key: Final[AnyKvKeyT_co]  # type: ignore[misc]

    @overload
    def __init__(self: _KeyBoundary[AnyKvKeyT_co], key: AnyKvKeyT_co) -> None: ...

    @overload
    def __init__(
        self: _KeyBoundary[KvKey[Unpack[Pieces]]], *pieces: Unpack[Pieces]
    ) -> None: ...

    def __init__(self, arg1: AnyKvKeyT_co | KvKeyPiece, *rest: KvKeyPiece) -> None:  # type: ignore[misc]
        key: AnyKvKey
        if is_any_kv_key(arg1):
            key = arg1
        else:
            key = KvKey(cast(KvKeyPiece, arg1), *rest)
        object.__setattr__(self, "key", key)

    @property
    def key_option(self) -> Some[AnyKvKeyT_co]:
        return Some(self.key)

    def range_start(self, start: StartT) -> KvKeyRange[StartT, Self]:  # type: ignore[type-var,unused-ignore]  # type-var because Self is not allowed to be the StopT type (but works anyway). unused-ignore because 3.9 does not detect the Self error.
        return KvKeyRange(start, self)  # type: ignore[type-var,unused-ignore]

    def range_stop(self, stop: StopT) -> KvKeyRange[Self, StopT]:  # type: ignore[type-var,unused-ignore]
        return KvKeyRange(self, stop)  # type: ignore[type-var,unused-ignore]

    def __repr__(self) -> str:
        if isinstance(self.key, KvKey):
            key = repr(tuple(self.key))[1:-1]
        else:
            key = repr(self.key)
        return f"{type(self).__name__}({key})"


class Include(_KeyBoundary[AnyKvKeyT_co]):
    """KvKeyRange boundary that includes its key in the range."""

    if TYPE_CHECKING:
        # For some reason mypy only infers types of Pieces using new not init
        @overload
        def __new__(cls, key: AnyKvKeyT, /) -> Include[AnyKvKeyT]: ...  # type: ignore[overload-overlap]

        @overload
        def __new__(cls, *pieces: Unpack[Pieces]) -> Include[KvKey[Unpack[Pieces]]]: ...

        def __new__(cls, *args: Any) -> Self: ...  # type: ignore[misc]

    def range(self) -> KvKeyRange[Self, Self]:
        """Get a key range including only this Include's key."""
        return KvKeyRange(self, self)


class IncludePrefix(_KeyBoundary[AnyKvKeyT_co]):
    """KvKeyRange boundary that includes keys prefixed by its key in the range."""

    if TYPE_CHECKING:
        # For some reason mypy only infers types of Pieces using new not init
        @overload
        def __new__(cls, key: AnyKvKeyT, /) -> IncludePrefix[AnyKvKeyT]: ...  # type: ignore[overload-overlap]

        @overload
        def __new__(
            cls, *pieces: Unpack[Pieces]
        ) -> IncludePrefix[KvKey[Unpack[Pieces]]]: ...

        def __new__(cls, *args: Any) -> Self: ...  # type: ignore[misc]

    @override
    def range_stop(self, stop: StopT) -> KvKeyRange[Include[AnyKvKeyT_co], StopT]:  # type: ignore[override]
        return KvKeyRange(Include(self.key), stop)

    def range(self) -> KvKeyRange[Include[AnyKvKeyT_co], Self]:
        """
        Create a key range that includes all child keys of this object's key.

        Examples
        --------
        >>> r = IncludePrefix('a', 1).range()
        >>> r
        KvKeyRange(start=Include('a', 1), stop=IncludePrefix('a', 1))
        >>> assert ('a', 0, 99) not in r
        >>> assert ('a', 1) in r
        >>> assert ('a', 1, 0) in r
        >>> assert ('a', 1, 99) in r
        >>> assert ('a', 2, 0) not in r
        """
        return KvKeyRange(Include(self.key), self)


class Exclude(_KeyBoundary[AnyKvKeyT_co]):
    """KvKeyRange boundary that excludes its key from the range."""

    if TYPE_CHECKING:
        # For some reason mypy only infers types of Pieces using new not init
        @overload
        def __new__(cls, key: AnyKvKeyT, /) -> Exclude[AnyKvKeyT]: ...  # type: ignore[overload-overlap]

        @overload
        def __new__(cls, *pieces: Unpack[Pieces]) -> Exclude[KvKey[Unpack[Pieces]]]: ...

        def __new__(cls, *args: Any) -> Self: ...  # type: ignore[misc]

    def range(self) -> KvKeyRange[Self, Self]:
        """Get an empty key range that excludes this Exclude's key at the boundaries."""
        return KvKeyRange(self, self)


@dataclass(frozen=True, **slots_if310())
class IncludeAll:
    """KvKeyRange boundary that includes any key in the range."""

    def __new__(cls) -> Self:
        # Always return the same unique instance from IncludeAll()
        instance = object.__new__(cls)

        def __new__(cls: type[IncludeAll]) -> IncludeAll:
            return instance

        cls.__new__ = __new__  # type: ignore[method-assign,assignment]
        return instance

    @property
    def key_option(self) -> Nothing:
        return Nothing()

    def range_start(self, start: StartT) -> KvKeyRange[StartT, Self]:
        return KvKeyRange(start, self)

    def range_stop(self, stop: StopT) -> KvKeyRange[Self, StopT]:
        return KvKeyRange(self, stop)

    @classmethod
    def range(cls) -> KvKeyRange[IncludeAll, IncludeAll]:
        """Get a key range that includes all keys."""
        return KvKeyRange(IncludeAll(), IncludeAll())


StartBoundary: TypeAlias = Union[
    IncludeAll, Include[AnyKvKeyT_co], Exclude[AnyKvKeyT_co]
]
StopBoundary: TypeAlias = Union[
    IncludeAll, Include[AnyKvKeyT], IncludePrefix[AnyKvKeyT], Exclude[AnyKvKeyT]
]

if TYPE_CHECKING:
    StartT = TypeVar(
        "StartT", bound=StartBoundary, default=StartBoundary, covariant=False
    )
    StartT_co = TypeVar(
        "StartT_co", bound=StartBoundary, default=StartBoundary, covariant=True
    )
    StopT = TypeVar("StopT", bound=StopBoundary, default=StopBoundary, covariant=False)
    StopT_co = TypeVar(
        "StopT_co", bound=StopBoundary, default=StopBoundary, covariant=True
    )
else:
    StartT = TypeVar("StartT", bound=StartBoundary, covariant=False)
    StopT = TypeVar("StopT", bound=StopBoundary, covariant=False)
    StartT_co = TypeVar("StartT_co", bound=StartBoundary, covariant=True)
    StopT_co = TypeVar("StopT_co", bound=StopBoundary, covariant=True)


@functools.total_ordering
@dataclass(frozen=True, eq=False, **slots_if310())
class KvKeyRange(KvKeyRangeEncodable, Generic[StartT_co, StopT_co]):
    """
    A range of KV key values, bounded by a start and end positions.

    Examples
    --------
    >>> key_range = Include('a', 0).range_stop(Exclude('a', 10))
    >>> key_range
    KvKeyRange(start=Include('a', 0), stop=Exclude('a', 10))
    >>> KvKey('a', 0) in key_range
    True
    >>> KvKey('a', 10) in key_range
    False
    >>> KvKey('a', 10) in key_range.range_stop(Include('a', 10))
    True
    """

    start: Final[StartT_co] = field(default=cast(StartT_co, IncludeAll()))  # type: ignore[misc]
    stop: Final[StopT_co] = field(default=cast(StopT_co, IncludeAll()))  # type: ignore[misc]
    _packed: tuple[bytes, bytes] | None = field(
        default=None, init=False, repr=False, hash=False, compare=False
    )

    def __post_init__(self) -> None:
        # Types don't allow IncludePrefix as start because it has the same
        # effect as Include at the start position. But if it is used, we
        # normalise it.
        if isinstance(self.start, IncludePrefix):
            object.__setattr__(self, "start", Include(self.start.key))

        if not isinstance(self.start, (IncludeAll, Include, Exclude)):
            raise TypeError("start must be IncludeAll, Include or Exclude")
        if not isinstance(self.stop, (IncludeAll, Include, IncludePrefix, Exclude)):
            raise TypeError(
                "stop must be IncludeAll, Include, IncludePrefix or Exclude"
            )

    def range_start(self, start: StartT) -> KvKeyRange[StartT, StopT_co]:
        """Get a key range with this range's stop and the provided `start`."""
        return KvKeyRange(start, self.stop)

    def range_stop(self, stop: StopT) -> KvKeyRange[StartT_co, StopT]:
        """Get a key range with this range's start and the provided `stop`."""
        return KvKeyRange(self.start, stop)

    @override
    def kv_key_range_bytes(self) -> tuple[bytes, bytes]:
        if (packed := self._packed) is not None:
            return packed

        options = PackKeyRangeOptions()

        start = self.start
        if isinstance(start, Include):
            options["start"] = start.key
        elif isinstance(start, Exclude):
            options["start"] = start.key
            options["exclude_start"] = True
        else:
            assert isinstance(start, IncludeAll)
            options["start"] = ()

        stop = self.stop
        if isinstance(stop, Include):
            options["end"] = stop.key
            options["exclude_end"] = False
        elif isinstance(stop, Exclude):
            options["end"] = stop.key
        elif isinstance(stop, IncludePrefix):
            options["prefix"] = stop.key
        else:
            assert isinstance(stop, IncludeAll)

        object.__setattr__(self, "_packed", (packed := pack_key_range(**options)))
        return packed

    def __contains__(self, x: object, /) -> bool:
        if not is_any_kv_key(x):
            return False
        packed = pack_key(x)
        packed_start, packed_stop = pack_key_range(self)
        return packed_start <= packed < packed_stop

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, KvKeyRange):
            return NotImplemented
        return self.kv_key_range_bytes() == value.kv_key_range_bytes()

    def __hash__(self) -> int:
        return hash(self.kv_key_range_bytes())

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, KvKeyRange):
            return NotImplemented
        return self.kv_key_range_bytes() < other.kv_key_range_bytes()



from __future__ import annotations

from abc import ABCMeta
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Final
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Protocol
from typing import TypeVar
from typing import Union
from typing import cast
from typing import overload
from typing import runtime_checkable

from denokv._pycompat.dataclasses import slots_if310

if TYPE_CHECKING:
    from typing_extensions import Never
    from typing_extensions import ParamSpec
    from typing_extensions import Self
    from typing_extensions import TypeAlias
    from typing_extensions import TypeGuard
    from typing_extensions import TypeIs

    P = ParamSpec("P")


@runtime_checkable
class AnySuccess(Protocol, metaclass=ABCMeta):
    def _AnySuccess_marker(self, no_call: Never) -> Never: ...


@runtime_checkable
class AnyFailure(Protocol, metaclass=ABCMeta):
    def _AnyFailure_marker(self, no_call: Never) -> Never: ...


T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
E = TypeVar("E")
E_co = TypeVar("E_co", covariant=True)
U = TypeVar("U")
V = TypeVar("V")
F = TypeVar("F", bound=Callable[..., Any])


def doc_from(source: object) -> Callable[[F], F]:
    def copy_docstring(fn: F) -> F:
        try:
            fn_name = fn.__name__
            t_fn = getattr(source, fn_name)
            t_fn_doc = t_fn.__doc__
            assert (
                getattr(fn, "__doc__", None) is None
            ), "target already has a __doc__ value"
            fn.__doc__ = t_fn_doc
        except Exception as e:
            raise ValueError(
                f"Failed to copy docstring from source {source!r} for function {fn!r}"
            ) from e
        return fn

    return copy_docstring


class OptionMethods(Iterable[T_co], Protocol[T_co]):
    """Methods implemented by Some and Nothing."""

    def or_raise(
        self, exc: Callable[P, BaseException], *exc_args: P.args, **exc_kwargs: P.kwargs
    ) -> T_co:
        """
        Return the value of this Some or throw an exception if this is Nothing.

        If exc is an exception type and no arguments are provided, the type is
        passed a string message indicating that the option is Nothing.

        Notes
        -----
        - Use `value_or()` / `value_or_else()` to return a default instead of
          raising.
        - Use the `value` property to get the value while raising a TypeError if
          the Option is Nothing.

        Examples
        --------
        >>> Some(1).or_raise(AssertionError)
        1
        >>> Nothing().or_raise(AssertionError)
        Traceback (most recent call last):
        AssertionError: attempted to access value from Nothing
        >>> Nothing().or_raise(ValueError, 'No values were provided')
        Traceback (most recent call last):
        ValueError: No values were provided
        """

    def filter(
        self, check: type[U] | Callable[[T_co], bool] | None = None
    ) -> Option[T_co]:
        """
        Return Some if its value passes a check, otherwise Nothing.

        If no check is provided, values are kept if they are True according to
        `bool()`.

        This method cannot narrow the return type, use
        [Options.filter](`denokv.result.Options.is_some_and`) to narrow the return
        type.

        Examples
        --------
        >>> Some('a').filter(int)
        Nothing()
        >>> Some(1).filter(int)
        Some(1)
        >>> Some(-1).filter(lambda x: x > 0)
        Nothing()
        >>> Some(1).filter(lambda x: x > 0)
        Some(1)
        """

    def flatten(self) -> Self:
        """
        Flatten a Some containing an Option into a single level.

        Examples
        --------
        >>> Some(Some(2)).flatten()
        Some(2)
        >>> Some(2).flatten()
        Some(2)
        >>> Some(Nothing()).flatten()
        Nothing()
        >>> Nothing().flatten()
        Nothing()
        """

    def inspect(self, fn: Callable[[T_co], None]) -> Self:
        """
        Call `fn(self.value)` only if this is Some, then return self.

        Examples
        --------
        >>> assert Some('Hello!').inspect(print) == Some('Hello!')
        Hello!
        >>> assert Nothing().inspect(print) == Nothing()
        """

    def map(self, fn: Callable[[T_co], U]) -> Option[U]:
        """
        Transform this Some's value with a function, otherwise Nothing.

        Examples
        --------
        >>> Some(2).map(lambda x: x * 2)
        Some(4)
        >>> Nothing().map(lambda x: x * 2)
        Nothing()
        """

    def map_or(self, default: U, fn: Callable[[T_co], V]) -> U | V:
        """
        Return `fn(this.value)` or default if this is Nothing.

        Examples
        --------
        >>> Some(2).map_or(-1, lambda x: x * 2)
        4
        >>> Nothing().map_or(-1, lambda x: x * 2)
        -1
        >>> Nothing().map_or('foo', lambda x: x * 2)
        'foo'
        """

    def map_or_else(
        self, default_fn: Callable[[], U], fn: Callable[[T_co], V]
    ) -> U | V:
        """
        Return `fn(this.value)` or `default_fn()` if this is Nothing.

        Examples
        --------
        >>> Some(2).map_or_else(lambda: -1, lambda x: x * 2)
        4
        >>> Nothing().map_or_else(lambda: -1, lambda x: x * 2)
        -1
        >>> Nothing().map_or_else(lambda: 'foo', lambda x: x * 2)
        'foo'
        """

    def ok_or(self, error: E) -> Result[T_co, E]:
        """
        Transform this into a Result of `Ok(self.value)` or `Err(error)` if Nothing.

        Examples
        --------
        >>> Some(1).ok_or('x')
        Ok(1)
        >>> Nothing().ok_or('x')
        Err('x')
        """

    def ok_or_else(self, fn: Callable[[], E]) -> Result[T_co, E]:
        """
        Transform this into a Result of `Ok(self.value)` or `Err(fn())` if Nothing.

        Examples
        --------
        >>> Some(1).ok_or_else(lambda: 'x')
        Ok(1)
        >>> Nothing().ok_or_else(lambda: 'x')
        Err('x')
        """

    def or_(self, other: Option[U]) -> Option[T_co | U]:
        """
        Return this if Some, otherwise other.

        Examples
        --------
        >>> Some(1).or_(Some(2))
        Some(1)
        >>> Some(1).or_(Nothing())
        Some(1)
        >>> Nothing().or_(Some(2))
        Some(2)
        """

    def or_else(self, fn: Callable[[], Option[U]]) -> Option[T_co | U]:
        """
        Return this if Some, otherwise the Option result of `fn()`.

        Examples
        --------
        >>> Some(1).or_else(lambda: Some(2))
        Some(1)
        >>> Some(1).or_else(lambda: Nothing())
        Some(1)
        >>> Nothing().or_else(lambda: Some(2))
        Some(2)
        """

    def value_or(self, default: U) -> T_co | U:
        """
        Return the value in this Some or default if this is Nothing.

        Examples
        --------
        >>> Some(1).value_or(2)
        1
        >>> Nothing().value_or(2)
        2
        """

    def value_or_else(self, fn: Callable[[], U]) -> T_co | U:
        """
        Return the value in this Some or the result of `fn()` if this is Nothing.

        Examples
        --------
        >>> Some(1).value_or_else(lambda: 2)
        1
        >>> Nothing().value_or_else(lambda: 2)
        2
        """

    def xor(self, other: Option[U]) -> Option[T_co | U]:
        """
        If just one of self and other is Some, return it, otherwise Nothing().

        Examples
        --------
        >>> Some(1).xor(Some(2))
        Nothing()
        >>> Some(1).xor(Nothing())
        Some(1)
        >>> Nothing().xor(Some(2))
        Some(2)
        """

    def zip(self, other: Option[U]) -> Option[tuple[T_co, U]]:
        """
        Pair this value with the value in other if both are Some.

        Examples
        --------
        >>> from denokv.result import Some, Nothing
        >>> Some(1).zip(Some(2))
        Some((1, 2))
        >>> Some(1).zip(Nothing())
        Nothing()
        >>> Nothing().zip(Some(2))
        Nothing()
        """

    def zip_with(self, other: Option[U], fn: Callable[[T_co, U], V]) -> Option[V]:
        """
        Call fn with the pair of this and the value in other if both are Some.

        Examples
        --------
        >>> Some('FF').zip_with(Some(16), int)
        Some(255)
        >>> Some('FF').zip_with(Nothing(), int)
        Nothing()
        >>> Nothing().zip_with(Some(16), int)
        Nothing()
        """


class Options(type):
    """
    Utility functions for Option (Some | None) types.

    Some and Nothing represent the presence or absence of a value as Some(_)
    and Nothing().
    """

    # mypy seems to ignore static methods if new is typed like this.
    if not TYPE_CHECKING:

        def __new__(self) -> Never:
            raise TypeError("cannot create instances of Options")

    @staticmethod
    def is_nothing(option: Option[T]) -> TypeIs[Nothing]:
        """
        Check if an Option is Nothing, narrowing its type.

        Example
        -------
        >>> Options.is_nothing(Some(1))
        False
        >>> Options.is_nothing(Nothing())
        True

        Notes
        -----
        This function is static because Python can't narrow the type of self
        with instance methods.
        """
        return isinstance(option, Nothing)

    @overload
    @staticmethod
    def is_nothing_or(option: Option[object], check: type[U]) -> TypeIs[Option[U]]: ...

    @overload
    @staticmethod
    def is_nothing_or(
        option: Option[T], check: Callable[[T], TypeIs[U]]
    ) -> TypeGuard[Option[U]]: ...

    # MyPy complains: "Overloaded function implementation does not accept all
    # possible arguments of signature 2". Seems fine to me...
    @staticmethod  # type: ignore[misc]
    def is_nothing_or(
        option: Option[T], check: type[U] | Callable[[T], TypeIs[U]]
    ) -> bool:
        """
        Narrow the type of an Option to Nothing, or Some whose value passes a test.

        Examples
        --------
        >>> is_negative = lambda x: x < 0
        >>> Options.is_nothing_or(Some(1), is_negative)
        False
        >>> Options.is_nothing_or(Some(-1), is_negative)
        True
        >>> Options.is_nothing_or(Nothing(), is_negative)
        True

        Notes
        -----
        This function is static because Python can't narrow the type of self
        with instance methods.
        """
        if isinstance(option, Nothing):
            return True
        if not isinstance(option, Some):
            raise TypeError(f"expected Some or Nothing, got: {option}")
        if isinstance(check, type):
            return isinstance(option.value, check)
        return check(option.value)

    @staticmethod
    def is_some(option: Option[T]) -> TypeIs[Some[T]]:
        """
        Check if an Option is Some, narrowing its type.

        Examples
        --------
        >>> Options.is_some(Some(1))
        True
        >>> Options.is_some(Nothing())
        False

        Notes
        -----
        This function is static because Python can't narrow the type of self
        with instance methods.
        """
        return isinstance(option, Some)

    @staticmethod
    def is_some_and(
        option: Option[T], check: Callable[[T], TypeIs[U]]
    ) -> TypeGuard[Some[U]]:
        """
        Narrow the type of an Option to Some whose value passes a test.

        Examples
        --------
        >>> is_positive = lambda x: x > 0

        >>> assert Options.is_some_and(Some(1), is_positive)
        >>> assert not Options.is_some_and(Some(-1), is_positive)
        >>> assert not Options.is_some_and(Nothing(), is_positive)

        Notes
        -----
        This function is static because Python can't narrow the type of self
        with instance methods.
        """
        return isinstance(option, Some) and check(option.value)

    @staticmethod
    def next(it: Iterable[T]) -> Option[T]:
        """
        Get the first value from an iterable as Some, or Nothing if it's empty.

        Examples
        --------
        >>> Options.next(())
        Nothing()
        >>> Options.next([3])
        Some(3)
        >>> it = iter([3])
        >>> Options.next(it), Options.next(it)
        (Some(3), Nothing())
        """
        try:
            return Some(next(iter(it)))
        except StopIteration:
            return Nothing()


@AnySuccess.register
@dataclass(frozen=True, init=False, **slots_if310())
class Some(Generic[T_co]):
    """
    A value — An Option representing the presence of a value.

    Examples
    --------
    >>> Some(2).value
    2
    >>> Some(2) == 2
    False
    >>> is_ok(Some(2))
    True
    >>> is_err(Some(2))
    False
    >>> is_err(Some(Exception()))
    False
    >>> is_ok(Some(Exception()))
    True
    """

    if TYPE_CHECKING:

        def _AnySuccess_marker(self, no_call: Never) -> Never: ...

    def __new__(cls, value: T_co) -> Some[T_co]:
        obj = object.__new__(Some)
        object.__setattr__(obj, "value", value)
        return obj

    value: Final[T_co]  # type: ignore[misc] # misc needed to ignore error: Final name must be initialized with a value. Final itself is needed because mypy under Python3.13 fail to detect Final via dataclass being frozen.
    """
    The value in this Some. Raises TypeError if accessed from Nothing.

    Examples
    --------
    >>> Some(1).value
    1
    >>> Nothing().value
    Traceback (most recent call last):
    TypeError: attempted to access value from Nothing
    """

    @doc_from(OptionMethods)
    def or_raise(
        self, exc: Callable[P, BaseException], *exc_args: P.args, **exc_kwargs: P.kwargs
    ) -> T_co:
        return self.value

    @doc_from(OptionMethods)
    def filter(
        self, check: type[U] | Callable[[T_co], bool] | None = None
    ) -> Option[T_co]:
        if isinstance(check, type):
            return self if isinstance(self.value, check) else Nothing()
        return self if (check or bool)(self.value) else Nothing()

    # Ignoring this is necessary to type this correctly, and seems fine in
    # practice. See https://stackoverflow.com/a/74567241/693728
    @overload
    def flatten(self: Some[Nothing]) -> Nothing: ...  # type: ignore[overload-overlap]
    @overload
    def flatten(self: Some[Some[U]]) -> Some[U]: ...
    @overload
    def flatten(self: Some[Option[U]]) -> Option[U]: ...  # type: ignore[overload-overlap]
    @overload
    def flatten(self) -> Self: ...

    def flatten(self: Some[Option[U] | T_co]) -> Option[U] | Some[T_co]:
        if isinstance(self.value, (Some, Nothing)):
            return self.value
        return cast(Some[T_co], self)

    doc_from(OptionMethods)(flatten)

    @doc_from(OptionMethods)
    def inspect(self, fn: Callable[[T_co], None]) -> Self:
        fn(self.value)
        return self

    @doc_from(OptionMethods)
    def map(self, fn: Callable[[T_co], U]) -> Some[U]:
        return Some(fn(self.value))

    @doc_from(OptionMethods)
    def map_or(self, default: U, fn: Callable[[T_co], V]) -> V:
        return fn(self.value)

    @doc_from(OptionMethods)
    def map_or_else(self, default_fn: Callable[[], U], fn: Callable[[T_co], V]) -> V:
        return fn(self.value)

    @doc_from(OptionMethods)
    def ok_or(self, error: E) -> Ok[T_co]:
        return Ok(self.value)

    @doc_from(OptionMethods)
    def ok_or_else(self, fn: Callable[[], E]) -> Ok[T_co]:
        return Ok(self.value)

    @doc_from(OptionMethods)
    def or_(self, other: Option[U]) -> Option[T_co]:
        return self

    @doc_from(OptionMethods)
    def or_else(self, fn: Callable[[], Option[U]]) -> Option[T_co]:
        return self

    @doc_from(OptionMethods)
    def value_or(self, default: U) -> T_co:
        return self.value

    @doc_from(OptionMethods)
    def value_or_else(self, fn: Callable[[], U]) -> T_co:
        return self.value

    def unzip(self: Some[tuple[U, V]]) -> tuple[Option[U], Option[V]]:
        """
        Transform an Option containing a pair into a pair of Options.

        Examples
        --------
        >>> Some((1, 2)).unzip()
        (Some(1), Some(2))
        >>> Nothing().unzip()
        (Nothing(), Nothing())
        >>> Some(42).unzip()
        Traceback (most recent call last):
        TypeError: attempted to unzip a Some not containing a pair
        """
        try:
            left, right = self.value
        except TypeError as e:
            raise TypeError("attempted to unzip a Some not containing a pair") from e
        return Some(left), Some(right)

    @doc_from(OptionMethods)
    def xor(self, other: Option[U]) -> Option[T_co | U]:
        if isinstance(other, Nothing):
            return self
        if isinstance(other, Some):
            return Nothing()
        return other

    @doc_from(OptionMethods)
    def zip(self, other: Option[U]) -> Option[tuple[T_co, U]]:
        if isinstance(other, Some):
            return Some((self.value, other.value))
        return Nothing()

    @doc_from(OptionMethods)
    def zip_with(self, other: Option[U], fn: Callable[[T_co, U], V]) -> Option[V]:
        if isinstance(other, Some):
            return Some(fn(self.value, other.value))
        return Nothing()

    def __repr__(self) -> str:
        return f"Some({self.value!r})"

    def __iter__(self) -> Iterator[T_co]:
        return iter((self.value,))


@AnyFailure.register
@dataclass(frozen=True, **slots_if310())
class Nothing:
    """
    No value — An Option representing the absence of a value.

    Examples
    --------
    >>> Nothing()
    Nothing()
    >>> Nothing() is Nothing()
    True
    >>> Nothing() == Some(None)
    False
    >>> is_err(Nothing())
    True
    >>> is_ok(Nothing())
    False
    """

    if TYPE_CHECKING:

        def _AnyFailure_marker(self, no_call: Never) -> Never: ...

    def __new__(cls) -> Self:
        instance = object.__new__(cls)

        def __new__(cls: type[Nothing]) -> Nothing:
            return instance

        Nothing.__new__ = __new__  # type: ignore[method-assign,assignment]
        return instance

    @doc_from(OptionMethods)
    def or_raise(
        self, exc: Callable[P, BaseException], *exc_args: P.args, **exc_kwargs: P.kwargs
    ) -> Never:
        if (
            not (exc_args or exc_kwargs)
            and isinstance(exc, type)
            and issubclass(exc, BaseException)
        ):
            exc_args = ("attempted to access value from Nothing",)  # type: ignore[assignment]
        raise exc(*exc_args, **exc_kwargs)

    @doc_from(OptionMethods)
    def filter(self, check: type[Any] | Callable[[Any], bool] | None = None) -> Nothing:
        return self

    @doc_from(OptionMethods)
    def flatten(self) -> Nothing:
        return self

    @doc_from(OptionMethods)
    def inspect(self, fn: Callable[[Any], None]) -> Nothing:
        return self

    @doc_from(OptionMethods)
    def map(self, fn: Callable[[Any], Any]) -> Nothing:
        return self

    @doc_from(OptionMethods)
    def map_or(self, default: U, fn: Callable[[Any], Any]) -> U:
        return default

    @doc_from(OptionMethods)
    def map_or_else(self, default_fn: Callable[[], U], fn: Callable[[Any], Any]) -> U:
        return default_fn()

    @doc_from(OptionMethods)
    def ok_or(self, error: E) -> Err[E]:
        return Err(error)

    @doc_from(OptionMethods)
    def ok_or_else(self, fn: Callable[[], E]) -> Err[E]:
        return Err(fn())

    @doc_from(OptionMethods)
    def or_(self, other: Option[U]) -> Option[U]:
        return other

    @doc_from(OptionMethods)
    def or_else(self, fn: Callable[[], Option[U]]) -> Option[U]:
        return fn()

    if not TYPE_CHECKING:

        @property
        def value(self) -> Never:
            raise TypeError("attempted to access value from Nothing")

    @doc_from(OptionMethods)
    def value_or(self, default: U) -> U:
        return default

    @doc_from(OptionMethods)
    def value_or_else(self, fn: Callable[[], U]) -> U:
        return fn()

    @doc_from(Some)
    def unzip(self) -> tuple[Nothing, Nothing]:
        return self, self

    @doc_from(OptionMethods)
    def xor(self, other: Option[U]) -> Option[T_co | U]:
        if isinstance(other, Some):
            return other
        return self

    @doc_from(OptionMethods)
    def zip(self, other: Option[U]) -> Nothing:
        return self

    @doc_from(OptionMethods)
    def zip_with(self, other: Option[U], fn: Callable[[Any, Any], Any]) -> Nothing:
        return self

    def __iter__(self) -> Iterator[Never]:
        return iter(())


Option: TypeAlias = Union[Some[T_co], Nothing]
"""The presence or absence of a value as Some(_) and Nothing()."""


class ResultMethods(Iterable[T_co], Protocol[T_co, E_co]):
    def and_(self, result: Result[U, E]) -> Result[U, E_co | E]:
        """
        Return result if this is Ok otherwise return this Err.

        Examples
        --------
        >>> assert Ok(2).and_(Ok(4)) == Ok(4)
        >>> assert Ok(2).and_(Ok(b'')) == Ok(b'')
        >>> assert Ok(2).and_(Err('x')) == Err('x')
        >>> assert Err('x').and_(Ok(4)) == Err('x')
        >>> assert Err('x').and_(Err('y')) == Err('x')
        """

    def and_then(self, fn: Callable[[T_co], Result[U, E]]) -> Result[U, E_co | E]:
        """
        Return the Result of `fn(self.value)` if this is Ok, otherwise return this Err.

        Examples
        --------
        >>> assert Ok(2).and_then(lambda x: Ok(x * 2)) == Ok(4)
        >>> assert Ok(2).and_then(lambda x: Ok([x])) == Ok([2])
        >>> assert Ok(2).and_then(lambda x: Err('x')) == Err('x')
        >>> assert Err('x').and_then(lambda x: Ok(x * 2)) == Err('x')
        """

    def error_or(self, default: U) -> E_co | U:
        """
        Return the Err's value, or default if this is Ok.

        Examples
        --------
        >>> assert Ok(1).error_or(2) == 2
        >>> assert Err('x').error_or(2) == 'x'
        """

    def error_or_else(self, fn: Callable[[], U]) -> E_co | U:
        """
        Return the Err's value, or call `fn()` if this is Ok.

        Examples
        --------
        >>> assert Ok(1).error_or_else(lambda: 2) == 2
        >>> assert Err('x').error_or_else(lambda: 2) == 'x'
        """

    # We have to type this quite broadly here, but can be more specific at the
    # Ok/Err implementations.
    def flatten(self) -> ResultMethods[object, object]:
        """
        Flatten an Ok containing a Result into a single Result.

        Ok not containing a Result are returned as is.

        Examples
        --------
        >>> assert Ok(Ok(2)).flatten() == Ok(2)
        >>> assert Ok(Err('x')).flatten() == Err('x')
        >>> assert Err('x').flatten() == Err('x')

        >>> assert Ok(Ok(Ok(2))).flatten() == Ok(Ok(2))

        >>> Ok(2).flatten()
        Ok(2)
        """

    def inspect(self, fn: Callable[[T_co], None]) -> Self:
        """
        Return as is after calling fn with the Ok's value only if this is Ok.

        Examples
        --------
        >>> assert Ok(2).inspect(lambda x: print('val:', x)) == Ok(2)
        val: 2
        >>> assert Err('x').inspect(lambda x: print('val:', x)) == Err('x')
        """

    def inspect_err(self, fn: Callable[[E_co], None]) -> Self:
        """
        Return as is after calling fn with the Err's error only if this is Err.

        Examples
        --------
        >>> assert Ok(2).inspect_err(lambda x: print('val:', x)) == Ok(2)
        >>> assert Err('x').inspect_err(lambda x: print('error:', x)) == Err('x')
        error: x
        """

    def map(self, fn: Callable[[T_co], U]) -> Result[U, E_co]:
        """
        Return `Ok(fn(self.value))` if this is Ok, or the Err as is.

        Examples
        --------
        >>> assert Ok(2).map(lambda x: x * 2) == Ok(4)
        >>> assert Err('x').map(lambda x: x * 2) == Err('x')
        """

    def map_or(self, default: U, fn: Callable[[T_co], U]) -> U:
        """
        Return `fn(self.value)` if this is Ok, or the default if this is Err.

        Examples
        --------
        >>> assert Ok(2).map_or(-1, lambda x: x * 2) == 4
        >>> assert Err('x').map_or(-1, lambda x: x * 2) == -1
        """

    def map_or_else(self, default: Callable[[E_co], U], fn: Callable[[T_co], U]) -> U:
        """
        Return `fn(self.value)` if this is Ok, or `default(self.error)` if this is Err.

        Examples
        --------
        >>> assert Ok(2).map_or_else(lambda e: -len(e), lambda x: x * 2) == 4
        >>> assert Err('x').map_or_else(lambda e: -len(e), lambda x: x * 2) == -1
        """

    def map_err(self, fn: Callable[[E_co], U]) -> Result[T_co, U]:
        """
        Return the Ok as is, or `Err(fn(self.error))` if this is Err.

        Examples
        --------
        >>> assert Ok(2).map_err(lambda e: f'Foo: {e}') == Ok(2)
        >>> assert Err('x').map_err(lambda e: f'Foo: {e}') == Err('Foo: x')
        """

    def ok(self) -> Option[T_co]:
        """
        Convert this Result to an Option of its value.

        Examples
        --------
        >>> Ok(2).ok()
        Some(2)
        >>> Err('fail').ok()
        Nothing()
        """

    def or_(self, default: Result[T, U]) -> Result[T_co | T, U]:
        """
        Return the Ok as-is, or the default Result if this is Err.

        Examples
        --------
        >>> assert Ok(1).or_(Ok(2)) == Ok(1)
        >>> assert Err('error a').or_(Ok(2)) == Ok(2)

        >>> assert Ok(1).or_(Err('error b')) == Ok(1)
        >>> assert Err('error a').or_(Err('error b')) == Err('error b')
        """

    def or_else(self, fn: Callable[[], Result[T, U]]) -> Result[T_co | T, U]:
        """
        Return the Ok as-is, or the Result from calling `fn()` if this is Err.

        Examples
        --------
        >>> assert Ok(1).or_else(lambda: Ok(2)) == Ok(1)
        >>> assert Err('error a').or_else(lambda: Ok(2)) == Ok(2)

        >>> assert Ok(1).or_else(lambda: Err('error b')) == Ok(1)
        >>> assert Err('error a').or_else(lambda: Err('error b')) == Err('error b')
        """

    def value_or(self, default: U) -> T_co | U:
        """
        Return the Ok's value, or default if this is Err.

        Examples
        --------
        >>> assert Ok(1).value_or(2) == 1
        >>> assert Err('x').value_or(2) == 2
        """

    def value_or_else(self, fn: Callable[[], U]) -> T_co | U:
        """
        Return the Ok's value, or call `fn()` if this is Err.

        Examples
        --------
        >>> assert Ok(1).value_or_else(lambda: 2) == 1
        >>> assert Err('x').value_or_else(lambda: 2) == 2
        """

    def __iter__(self) -> Iterator[T_co]:
        """
        Return an iterator containing the Ok's value or no values if this is Err.

        Examples
        --------
        >>> assert list(Ok(1)) == [1]
        >>> assert list(Err('error a')) == []
        """


class Results(ResultMethods[T_co, E_co]):
    """
    Utility functions for Result (Ok | Err) types.

    Ok and Err represent the presence of a value, or a reason the value could
    not be produced.

    Functions that may fail can return a Result rather than raising an Exception
    to signal failure.

    Examples
    --------
    >>> def parse(x: str) -> Result[int, ValueError]:
    ...     try:
    ...         return Ok(int(x))
    ...     except ValueError as e:
    ...         return Err(e)
    >>> parse('42')
    Ok(42)
    >>> parse('a4')
    Err(ValueError("invalid literal for int() with base 10: 'a4'"))
    """

    # mypy seems to ignore static methods if new is typed like this.
    if not TYPE_CHECKING:

        def __new__(self) -> Never:
            raise TypeError("cannot create instances of Results")

    @staticmethod
    def call(
        fn: Callable[P, T], *fn_args: P.args, **fn_kwargs: P.kwargs
    ) -> Result[T, Exception]:
        """
        Call a function and return Ok with its return value, or Err if it raises.

        Examples
        --------
        >>> Results.call(int, '42')
        Ok(42)
        >>> Results.call(int, 'a4')
        Err(ValueError("invalid literal for int() with base 10: 'a4'"))
        """
        try:
            return Ok(fn(*fn_args, **fn_kwargs))
        except Exception as e:
            return Err(e)

    def __init_subclass__(cls) -> None:
        module = globals()
        if "Ok" in module and "Err" in module:
            raise TypeError("cannot subclass Result")

    @staticmethod
    def is_ok(result: Result[T_co, Any]) -> TypeIs[Ok[T_co]]:
        return isinstance(result, Ok)

    # Note: It doesn't seem to be possible to use TypeIs as the return of
    # is_err_and. If we do, we have to make T object to satisfy the subtype
    # return requirement, and also mypy incorrectly over-narrows the non-matching
    # case to Ok, instead of keeping the non-matching Err.
    @staticmethod
    def is_ok_and(
        result: Result[T, object], check: Callable[[T], TypeIs[U]]
    ) -> TypeGuard[Ok[U]]:
        """Narrow the type of a result to Ok with a particular value type."""
        return isinstance(result, Ok) and check(result.value)

    @staticmethod
    def is_err(result: Result[Any, E_co]) -> TypeIs[Err[E_co]]:
        return isinstance(result, Err)

    # Note: It doesn't seem to be possible to use TypeIs as the return of
    # is_err_and. If we do, we have to make T object to satisfy the subtype
    # return requirement, and also mypy incorrectly over-narrows the non-matching
    # case to Ok, instead of keeping the non-matching Err.
    @staticmethod
    def is_err_and(
        result: Result[Any, T], check: Callable[[T], TypeIs[U]]
    ) -> TypeGuard[Err[U]]:
        """Narrow the type of a result to Err with a particular error type."""
        return isinstance(result, Err) and check(result.error)


@AnySuccess.register
@dataclass(frozen=True, **slots_if310())
class Ok(Generic[T_co]):
    if TYPE_CHECKING:

        def _AnySuccess_marker(self, no_call: Never) -> Never: ...

    value: Final[T_co]  # type: ignore[misc] # misc needed to ignore error: Final name must be initialized with a value. Final itself is needed because mypy under Python3.13 fail to detect Final via dataclass being frozen.
    """
    Access the Ok's value. Raises TypeError if this is Err.

    Examples
    --------
    >>> assert Ok(1).value == 1
    >>> Err('x').value
    Traceback (most recent call last):
    TypeError: attempted to access value from Err
    """

    @doc_from(ResultMethods)
    def and_(self, result: Result[U, E]) -> Result[U, E]:
        return result

    @doc_from(ResultMethods)
    def and_then(self, fn: Callable[[T_co], Result[U, E]]) -> Result[U, E]:
        return fn(self.value)

    if not TYPE_CHECKING:

        @property
        def error(self) -> Never:
            raise TypeError("attempted to access error from Ok")

    @doc_from(ResultMethods)
    def error_or(self, default: U) -> U:
        return default

    @doc_from(ResultMethods)
    def error_or_else(self, fn: Callable[[], U]) -> U:
        return fn()

    @overload
    def flatten(self: Ok[Ok[U]]) -> Ok[U]: ...
    @overload
    def flatten(self: Ok[Err[E]]) -> Err[E]: ...
    @overload
    def flatten(self: Ok[Result[U, E]]) -> Result[U, E]: ...
    @overload
    def flatten(self) -> Result[T_co, E]: ...

    def flatten(self: Ok[Result[U, E] | T_co]) -> Ok[U] | Err[E] | Result[T_co, E]:
        if isinstance(self.value, (Ok, Err)):
            return self.value
        return cast(Ok[T_co], self)

    doc_from(ResultMethods)(flatten)

    @doc_from(ResultMethods)
    def inspect(self, fn: Callable[[T_co], None]) -> Self:
        fn(self.value)
        return self

    @doc_from(ResultMethods)
    def inspect_err(self, fn: Callable[[E_co], None]) -> Self:
        return self

    @doc_from(ResultMethods)
    def map(self, fn: Callable[[T_co], U]) -> Ok[U]:
        return Ok(fn(self.value))

    @doc_from(ResultMethods)
    def map_or(self, default: U, fn: Callable[[T_co], U]) -> U:
        return fn(self.value)

    @doc_from(ResultMethods)
    def map_or_else(self, default: Callable[[E_co], U], fn: Callable[[T_co], U]) -> U:
        return fn(self.value)

    @doc_from(ResultMethods)
    def map_err(self, fn: Callable[[E_co], U]) -> Self:
        return self

    @doc_from(ResultMethods)
    def ok(self) -> Some[T_co]:
        return Some(self.value)

    @doc_from(ResultMethods)
    def or_(self, default: Result[T, U]) -> Result[T_co, U]:
        return self

    @doc_from(ResultMethods)
    def or_else(self, fn: Callable[[], Result[T, U]]) -> Result[T_co, U]:
        return self

    @doc_from(ResultMethods)
    def value_or(self, default: U) -> T_co:
        return self.value

    @doc_from(ResultMethods)
    def value_or_else(self, fn: Callable[[], U]) -> T_co:
        return self.value

    def __iter__(self) -> Iterator[T_co]:
        return iter((self.value,))

    def __repr__(self) -> str:
        return f"Ok({self.value!r})"


@AnyFailure.register
@dataclass(frozen=True, **slots_if310())
class Err(Generic[E_co]):
    if TYPE_CHECKING:

        def _AnyFailure_marker(self, no_call: Never) -> Never: ...

    error: Final[E_co]  # type: ignore[misc] # misc needed to ignore error: Final name must be initialized with a value. Final itself is needed because mypy under Python3.13 fail to detect Final via dataclass being frozen.
    """
    Access the Err's error. Raises TypeError if this is Ok.

    Examples
    --------
    >>> assert Err('x').error == 'x'
    >>> Ok(1).error
    Traceback (most recent call last):
    TypeError: attempted to access error from Ok
    """

    @doc_from(ResultMethods)
    def and_(self, result: Result[U, E]) -> Self:
        return self

    @doc_from(ResultMethods)
    def and_then(self, fn: Callable[[T_co], Result[U, E]]) -> Self:
        return self

    @doc_from(ResultMethods)
    def error_or(self, default: U) -> E_co:
        return self.error

    @doc_from(ResultMethods)
    def error_or_else(self, fn: Callable[[], U]) -> E_co:
        return self.error

    @doc_from(ResultMethods)
    def flatten(self) -> Self:
        return self

    @doc_from(ResultMethods)
    def inspect(self, fn: Callable[[T_co], None]) -> Self:
        return self

    @doc_from(ResultMethods)
    def inspect_err(self, fn: Callable[[E_co], None]) -> Self:
        fn(self.error)
        return self

    @doc_from(ResultMethods)
    def map(self, fn: Callable[[T_co], U]) -> Self:
        return self

    @doc_from(ResultMethods)
    def map_or(self, default: U, fn: Callable[[T_co], U]) -> U:
        return default

    @doc_from(ResultMethods)
    def map_or_else(self, default: Callable[[E_co], U], fn: Callable[[T_co], U]) -> U:
        return default(self.error)

    @doc_from(ResultMethods)
    def map_err(self, fn: Callable[[E_co], U]) -> Err[U]:
        return Err(fn(self.error))

    @doc_from(ResultMethods)
    def ok(self) -> Nothing:
        return Nothing()

    @doc_from(ResultMethods)
    def or_(self, default: Result[T_co, U]) -> Result[T_co, U]:
        return default

    @doc_from(ResultMethods)
    def or_else(self, fn: Callable[[], Result[T_co, U]]) -> Result[T_co, U]:
        return fn()

    if not TYPE_CHECKING:

        @property
        def value(self) -> Never:
            raise TypeError("attempted to access value from Err")

    @doc_from(ResultMethods)
    def value_or(self, x_default: U) -> U:
        return x_default

    @doc_from(ResultMethods)
    def value_or_else(self, fn: Callable[[], U]) -> U:
        return fn()

    def __iter__(self) -> Iterator[Never]:
        return iter(())

    def __repr__(self) -> str:
        return f"Err({self.error!r})"


Result: TypeAlias = Union[Ok[T_co], Err[E_co]]
"""Represents the presence of a value, or a reason the value could not be produced."""


def is_ok(result: object) -> TypeIs[AnySuccess]:
    """Check if a value is a [successful type](`denokv.result.AnySuccess`)."""
    return isinstance(result, AnySuccess)


def is_err(result: object) -> TypeIs[AnyFailure]:
    """Check if a value is a [failure type](`denokv.result.AnyFailure`)."""
    return isinstance(result, AnyFailure)



from __future__ import annotations

from typing import overload

from denokv._datapath_pb2 import AtomicWriteStatus
from denokv._datapath_pb2 import MutationType
from denokv._datapath_pb2 import ValueEncoding


@overload
def enum_name(enum_type: type[AtomicWriteStatus], value: AtomicWriteStatus) -> str: ...
@overload
def enum_name(enum_type: type[MutationType], value: MutationType) -> str: ...
@overload
def enum_name(enum_type: type[ValueEncoding], value: ValueEncoding) -> str: ...
def enum_name(
    enum_type: type, value: MutationType | ValueEncoding | AtomicWriteStatus
) -> str:
    return enum_type.Name(value)  # type: ignore[no-any-return,attr-defined]



from __future__ import annotations

import asyncio
import weakref
from base64 import urlsafe_b64decode
from base64 import urlsafe_b64encode
from binascii import unhexlify
from contextlib import AbstractAsyncContextManager
from dataclasses import dataclass
from dataclasses import field
from enum import Flag
from enum import auto
from os import environ
from types import TracebackType
from typing import TYPE_CHECKING
from typing import AsyncIterator
from typing import Awaitable
from typing import Callable
from typing import ClassVar
from typing import Final
from typing import Generic
from typing import Iterable
from typing import Literal
from typing import Protocol
from typing import Sequence
from typing import TypedDict
from typing import overload

import aiohttp
from fdb.tuple import unpack
from v8serialize import Decoder
from yarl import URL

from denokv import datapath
from denokv._datapath_pb2 import ReadRange
from denokv._datapath_pb2 import SnapshotRead
from denokv._datapath_pb2 import SnapshotReadOutput
from denokv._pycompat.dataclasses import slots_if310
from denokv._pycompat.typing import override
from denokv.asyncio import loop_time
from denokv.auth import ConsistencyLevel
from denokv.auth import DatabaseMetadata
from denokv.auth import EndpointInfo
from denokv.auth import MetadataExchangeDenoKvError
from denokv.auth import get_database_metadata
from denokv.backoff import Backoff
from denokv.backoff import ExponentialBackoff
from denokv.backoff import attempts
from denokv.datapath import AnyKvKey
from denokv.datapath import AnyKvKeyT
from denokv.datapath import AutoRetry
from denokv.datapath import DataPathError
from denokv.datapath import KvKeyEncodable
from denokv.datapath import KvKeyPiece
from denokv.datapath import KvKeyTuple
from denokv.datapath import ProtocolViolation
from denokv.datapath import SnapshotReadResult
from denokv.datapath import is_kv_key_tuple
from denokv.datapath import pack_key
from denokv.datapath import parse_protobuf_kv_entry
from denokv.datapath import read_range_multi
from denokv.datapath import read_range_single
from denokv.errors import InvalidCursor
from denokv.kv_keys import KvKey
from denokv.result import Err
from denokv.result import Ok
from denokv.result import Result

if TYPE_CHECKING:
    from typing_extensions import Self
    from typing_extensions import TypeAlias
    from typing_extensions import TypeVar
    from typing_extensions import TypeVarTuple
    from typing_extensions import Unpack

    T = TypeVar("T", default=object)
    # Note that the default arg doesn't seem to work with MyPy yet. The
    # DefaultKvKey alias is what this should behave as when defaulted.
    Pieces = TypeVarTuple("Pieces", default=Unpack[tuple[KvKeyPiece, ...]])
else:
    from typing import TypeVar

    T = TypeVar("T")
    Unpack = tuple  # hack to support py39 at runtime w/o typing_extensions
    Pieces = TypeVar("Pieces")  # hack to support py39 at runtime w/o typing_extensions

SAFE_FLOAT_INT_RANGE: Final = range(-(2**53 - 1), 2**53)  # 2**53 - 1 is max safe

CursorFormatType: TypeAlias = Callable[["ListContext"], "AnyCursorFormat"]


class KvListOptions(TypedDict, total=False):
    """Keyword arguments of `Kv.list()`."""

    limit: int | None
    cursor: str | None
    reverse: bool | None  # order asc/desc?
    consistency: ConsistencyLevel | None
    batch_size: int | None
    cursor_format_type: CursorFormatType | None


@dataclass(frozen=True, **slots_if310())
class KvEntry(Generic[AnyKvKeyT, T]):
    """A value read from the Deno KV database, along with its key and version."""

    key: AnyKvKeyT
    value: T
    versionstamp: VersionStamp


@dataclass(frozen=True, **slots_if310())
class ListKvEntry(KvEntry[AnyKvKeyT, T]):
    """
    A value read from the Deno KV database with a list operation.

    In addition to a normal [KvEntry's] key, value and version, [ListKvEntry]
    provides a [cursor] that can be used to start another [Kv.list()] starting
    from the value after this.
    """

    listing: ListContext = field(repr=False)

    @property
    def cursor(self) -> str:
        result = self.listing.cursor_format.get_cursor_for_key(self.key)
        if isinstance(result, Err):
            raise result.error  # should never occur with entries from Kv.list()
        return result.value


class VersionStamp(bytes):
    r"""
    A 20-hex-char / (10 byte) version identifier.

    This value represents the relative age of a KvEntry. A VersionStamp that
    compares larger than another is newer.

    Examples
    --------
    >>> VersionStamp(0xff << 16)
    VersionStamp('00000000000000ff0000')
    >>> int(VersionStamp('000000000000000000ff'))
    255
    >>> bytes(VersionStamp('00000000000000ff0000'))
    b'\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00'
    >>> VersionStamp(b'\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00')
    VersionStamp('00000000000000ff0000')
    >>> isinstance(VersionStamp(0), bytes)
    True
    >>> str(VersionStamp(0xff << 16))
    '00000000000000ff0000'
    """

    RANGE: ClassVar = range(0, 2**80)

    def __new__(cls, value: str | bytes | int) -> Self:
        if isinstance(value, int):
            if value not in VersionStamp.RANGE:
                raise ValueError("value not in range for 80-bit unsigned int")
            # Unlike most others, versionstamp uses big-endian as it needs to
            # sort lexicographically as bytes.
            value = value.to_bytes(length=10, byteorder="big")
        if isinstance(value, str):
            try:
                value = unhexlify(value)
            except Exception:
                value = b""
            if len(value) != 10:
                raise ValueError("value is not a 20 char hex string")
        else:
            if len(value) != 10:
                raise ValueError("value is not 10 bytes long")
        return bytes.__new__(cls, value)

    def __index__(self) -> int:
        return int.from_bytes(self, byteorder="big")

    def __bytes__(self) -> bytes:
        return self[:]

    def __str__(self) -> str:
        return self.hex()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"


@dataclass(frozen=True, **slots_if310())
class KvU64:
    """
    An special int value that supports operations like `sum`, `max`, and `min`.

    Notes
    -----
    This type is not an int subtype to avoid it being mistakenly flattened into
    a regular int and loosing its special meaning when written back to the DB.

    Examples
    --------
    >>> KvU64(bytes([0, 0, 0, 0, 0, 0, 0, 0]))
    KvU64(0)
    >>> KvU64(bytes([1, 0, 0, 0, 0, 0, 0, 0]))
    KvU64(1)
    >>> KvU64(bytes([1, 1, 0, 0, 0, 0, 0, 0]))
    KvU64(257)
    >>> KvU64(2**64 - 1)
    KvU64(18446744073709551615)
    >>> KvU64(2**64)
    Traceback (most recent call last):
    ...
    ValueError: value not in range for 64-bit unsigned int
    >>> KvU64(-1)
    Traceback (most recent call last):
    ...
    ValueError: value not in range for 64-bit unsigned int
    """

    RANGE: ClassVar = range(0, 2**64)
    value: int

    def __init__(self, value: bytes | int) -> None:
        if isinstance(value, bytes):
            if len(value) != 8:
                raise ValueError("value must be a 8 bytes")
            value = int.from_bytes(value, byteorder="little")
        elif isinstance(value, int):
            if value not in KvU64.RANGE:
                raise ValueError("value not in range for 64-bit unsigned int")
        else:
            raise TypeError("value must be 8 bytes or a 64-bit unsigned int")
        object.__setattr__(self, "value", value)

    def __index__(self) -> int:
        return self.value

    def __bytes__(self) -> bytes:
        return self.to_bytes()

    def to_bytes(self) -> bytes:
        return self.value.to_bytes(8, byteorder="little")

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.value})"


@dataclass(frozen=True, **slots_if310())
class EndpointSelector:
    # Right now this is very simple, which is fine for the local SQLite-backed
    # Deno KV server, but for distributed Deno KV we need to support selecting
    # endpoints based on latency, so this can be stateful.
    meta: DatabaseMetadata

    def __post_init__(self) -> None:
        # Require at least one strongly-consistent endpoint. Note that this is
        # a requirement of the KV Connect spec and we enforce it when fetching
        # metadata, so this requirement should only be broken as a result of a
        # programmer error.
        if not any(
            ep.consistency is ConsistencyLevel.STRONG for ep in self.meta.endpoints
        ):
            raise ValueError(f"no endpoint has {ConsistencyLevel.STRONG} consistency")

    def get_endpoints(self, consistency: ConsistencyLevel) -> Sequence[EndpointInfo]:
        return [
            e
            for e in self.meta.endpoints
            if consistency is ConsistencyLevel.EVENTUAL
            or e.consistency is ConsistencyLevel.STRONG
        ]

    def get_endpoint(self, consistency: ConsistencyLevel) -> EndpointInfo:
        candidates = self.get_endpoints(consistency)
        # DatabaseMetadata is required by the Metadata Exchange spec to have at
        # least one strongly consistent endpoint.
        assert len(candidates) > 0
        return candidates[0]


@dataclass(frozen=True, **slots_if310())
class CachedValue(Generic[T]):
    fresh_until: float
    value: T

    def is_fresh(self, time: float) -> bool:
        return time < self.fresh_until


AuthenticatorFn: TypeAlias = Callable[
    [], Awaitable["Result[DatabaseMetadata, MetadataExchangeDenoKvError]"]
]
"""
The type of a function that connects to a KV database and returns metadata.

The metadata contains the server URLs and access tokens needed to query the KV
database.
"""


@dataclass(frozen=True, **slots_if310())
class KvCredentials:
    server_url: URL
    access_token: str


@dataclass
class Authenticator:
    """
    Authenticates with a KV database server and returns its metadata.

    Notes
    -----
    This is an [AuthenticationFn](`denokv.kv.AuthenticationFn`).
    """

    session: aiohttp.ClientSession
    retry_delays: Backoff
    credentials: KvCredentials

    async def __call__(self) -> Result[DatabaseMetadata, MetadataExchangeDenoKvError]:
        result: Result[DatabaseMetadata, MetadataExchangeDenoKvError] | None = None
        credentials = self.credentials
        for delay in attempts(self.retry_delays):
            result = await get_database_metadata(
                session=self.session,
                server_url=credentials.server_url,
                access_token=credentials.access_token,
            )
            if isinstance(result, Err) and result.error.retryable:
                await asyncio.sleep(delay)
                continue
            return result
        else:  # backoff timed out
            assert isinstance(result, Err)
            return result


def _cached_database_metadata(value: DatabaseMetadata) -> CachedValue[DatabaseMetadata]:
    return CachedValue(fresh_until=loop_time(wall_time=value.expires_at), value=value)


def normalize_key(key: KvKeyTuple, *, bigints: bool = False) -> KvKeyTuple:
    """
    Coerce key types for JavaScript compatibility.

    Encoding of `int` values depends on the `bigints` argument:
    - by default and with `bigints=False` `int` is encoded as `float`, which is
        how JavaScript encodes `number`.
    - with `bigints=True` `int` is encoded as JavaScript encodes `BigInt` and
      `float` is encoded as JavaScript encodes `number`.

    [SAFE_FLOAT_INT_RANGE](`denokv.kv.SAFE_FLOAT_INT_RANGE`)

    Raises
    ------
    ValueError
        If `bigints=False` and an int value is outside [SAFE_FLOAT_INT_RANGE].

    Examples
    --------
    The `key` can contain int, but ints are packed as float because JavaScript
    int are floats.

    >>> normalize_key((1, 2.0), bigints=False)
    (1.0, 2.0)

    Pass `bigints=True` to encode int as JavaScript encodes BigInt:

    >>> normalize_key((1, 2.0), bigints=True)
    (1, 2.0)
    """
    if bigints or not any(isinstance(part, int) for part in key):
        return key
    normalised = list(key)
    for i, val in enumerate(normalised):
        # bool is an int subtype!
        if not (isinstance(val, int) and not isinstance(val, bool)):
            continue
        if val not in SAFE_FLOAT_INT_RANGE:
            raise ValueError(
                f"int value is too large to be losslessly normalized to a float: {val}."
            )
        normalised[i] = float(val)
    return tuple(normalised)


@dataclass(init=False)
class DatabaseMetadataCache:
    authenticator: AuthenticatorFn
    current: CachedValue[DatabaseMetadata] | None
    pending: (
        asyncio.Task[Result[CachedValue[DatabaseMetadata], MetadataExchangeDenoKvError]]
        | None
    )

    def __init__(
        self,
        *,
        initial: DatabaseMetadata | None = None,
        authenticator: AuthenticatorFn,
    ) -> None:
        # Can start as None, in which case reload happens on first access.
        self.current = _cached_database_metadata(initial) if initial else None
        self.pending = None
        self.authenticator = authenticator

    async def get(
        self, now: float | None = None
    ) -> Result[DatabaseMetadata, MetadataExchangeDenoKvError]:
        if now is None:
            now = loop_time()
        current = self.current
        if current is not None and current.is_fresh(now):
            return Ok(current.value)

        pending_task = self.pending
        if pending_task is None:
            self.pending = pending_task = asyncio.create_task(self.reload())
        pending = await pending_task

        # The first caller to await pending can handle swapping pending to current
        if self.pending is pending_task:
            self.pending = None
            # We don't cache error values. We expect that the auth fn called by
            # reload() is doing its own retry and backoff, and the Kv call that
            # triggers get() (e.g. Kv.get()) will also do retry with backoff on
            # retry-able failed auth.
            # TODO: It'd probably make sense for the auth fn to also have a rate
            #   limit on attempts after non-retry-able errors.
            if isinstance(pending, Ok):
                self.current = pending.value

        if isinstance(pending, Err):
            return pending
        return Ok(pending.value.value)

    async def reload(
        self,
    ) -> Result[CachedValue[DatabaseMetadata], MetadataExchangeDenoKvError]:
        result = await self.authenticator()
        if isinstance(result, Err):
            return result
        return Ok(_cached_database_metadata(result.value))

    def purge(self, metadata: DatabaseMetadata) -> None:
        # Only purge the cached version if the purged version is the current,
        # otherwise an async task operating on an expired previous version could
        # incorrectly expire a just-fetched fresh version.
        if self.current and self.current.value is metadata:
            self.current = None


class KvFlags(Flag):
    """Options that can be enabled/disabled to affect [Kv](`denokv.kv.Kv`) behaviour."""

    NoFlag = 0
    IntAsNumber = auto()
    """
    Treat `int` as `float` in KV keys.

    This causes ints to behave like JavaScript number values, as JavaScript uses
    floating-point numbers for integer literals. In JavaScript, Deno KV also
    supports `BigInt` keys. To use Python `int` like JavaScript `BigInt` keys,
    either don't use this flag (and manually convert to float only where
    needed), or pass keys as [KvKey](`denokv.kv.KvKey`) values, which maintain
    `int` values as-is, causing them to act like JavaScript `BigInt`.
    """


DEFAULT_KV_FLAGS: Final = KvFlags.IntAsNumber


@dataclass(init=False)
class Kv(AbstractAsyncContextManager["Kv", None]):
    """
    Interface to perform requests against a Deno KV database.

    [DEFAULT_KV_FLAGS]: `denokv.kv.DEFAULT_KV_FLAGS`

    Parameters
    ----------
    flags
        Enable/disable flags that change Kv behaviour. Default: [DEFAULT_KV_FLAGS]
    """

    session: aiohttp.ClientSession
    retry_delays: Backoff
    metadata_cache: DatabaseMetadataCache
    v8_decoder: Decoder
    flags: KvFlags

    def __init__(
        self,
        session: aiohttp.ClientSession,
        auth: AuthenticatorFn,
        retry: Backoff | None = None,
        v8_decoder: Decoder | None = None,
        flags: KvFlags | None = None,
    ) -> None:
        self.session = session
        self.metadata_cache = DatabaseMetadataCache(authenticator=auth)
        self.retry_delays = ExponentialBackoff() if retry is None else retry
        self.v8_decoder = v8_decoder or Decoder()
        self.flags = KvFlags.IntAsNumber if flags is None else flags

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        await self.aclose()

    @property
    def closed(self) -> bool:
        return self.session.closed

    async def aclose(self) -> None:
        if self.closed:
            return
        await self._aclose(self.session)

    @classmethod
    async def _aclose(cls, session: aiohttp.ClientSession) -> None:
        await session.close()

    @classmethod
    def _finalize(cls, session: aiohttp.ClientSession) -> None | asyncio.Future[None]:
        if session._loop.is_running():
            return session._loop.create_task(
                cls._aclose(session), name="denokv.Kv.create_finalizer"
            )
        else:
            return session._loop.run_until_complete(cls._aclose(session))

    def create_finalizer(self) -> weakref.finalize:
        """
        Automatically close the instance when it goes out of scope, or at exit.

        This creates a Finalizer (`weakref.finalize`) that closes the Kv
        instance automatically when garbage collected, or when Python exits.

        If the event loop of the Kv's session is running, the finalizer returns,
        an `asyncio.Task` that closes the Kv instance. Otherwise it runs the
        session's loop to close it and returns None after it's closed.

        Notes
        -----
        It's recommended to close Kv instances explicitly using async context
        manager blocks, but Finalizers can be used in situations where a context
        manager is not practical, like in an interactive environment.

        `open_kv()` automatically creates a Finalizer in interactive sessions.
        """
        return weakref.finalize(self, self._finalize, self.session)

    def _prepare_key(self, key: AnyKvKeyT) -> AnyKvKeyT:
        if self.flags & KvFlags.IntAsNumber and not isinstance(key, KvKeyEncodable):
            return normalize_key(key, bigints=False)  # type: ignore[return-value]
        return key

    # get(x), get(x, y), get(keys=[a, b, c])
    @overload
    async def get(
        self,
        /,
        *,
        keys: Iterable[AnyKvKeyT],
        consistency: ConsistencyLevel = ConsistencyLevel.STRONG,
    ) -> tuple[tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None], ...]: ...

    @overload
    async def get(
        self,
        key1: AnyKvKeyT,
        key2: AnyKvKeyT,
        /,
        *keys: AnyKvKeyT,
        consistency: ConsistencyLevel = ConsistencyLevel.STRONG,
    ) -> tuple[tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None], ...]: ...

    @overload
    async def get(
        self,
        key: AnyKvKeyT,
        /,
        *,
        consistency: ConsistencyLevel = ConsistencyLevel.STRONG,
    ) -> tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None]: ...

    async def get(
        self,
        *args: AnyKvKeyT,
        keys: Iterable[AnyKvKeyT] | None = None,
        consistency: ConsistencyLevel = ConsistencyLevel.STRONG,
    ) -> (
        tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None]
        | tuple[tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None], ...]
    ):
        """Get the value of one or more known keys from the database."""
        if keys is not None:
            if len(args) > 0:
                raise TypeError("cannot use positional keys and keys keyword argument")
            args = tuple(keys)
            return_unwrapped = False
        else:
            if len(args) == 0:
                raise TypeError("at least one key argument must be passed")
            return_unwrapped = len(args) == 1

        args = tuple(self._prepare_key(key) for key in args)
        ranges = [read_range_single(key) for key in args]
        snapshot_read_result = await self._snapshot_read(
            ranges, consistency=consistency
        )
        if isinstance(snapshot_read_result, Err):
            raise snapshot_read_result.error
        result, endpoint = snapshot_read_result.value

        assert len(args) == len(ranges)

        results: list[tuple[AnyKvKeyT, KvEntry[AnyKvKeyT] | None]] = []
        decoder = self.v8_decoder
        for key, in_range, range in zip(args, ranges, result.ranges):
            if len(range.values) == 0:
                results.append((key, None))
                continue
            if len(range.values) != 1:
                raise ProtocolViolation(
                    f"Server responded with {len(range.values)} values to "
                    f"read for key {key!r} with limit 1",
                    data=result,
                    endpoint=endpoint,
                )
            raw_kv_entry = range.values[0]
            if raw_kv_entry.key != in_range.start:
                raise ProtocolViolation(
                    f"Server responded to read for exact key "
                    f"{in_range.start!r} with key {raw_kv_entry.key!r}",
                    data=result,
                    endpoint=endpoint,
                )

            parse_result = parse_protobuf_kv_entry(
                raw_kv_entry, v8_decoder=decoder, le64_type=KvU64
            )
            if isinstance(parse_result, Err):
                raise ProtocolViolation(
                    f"Server responded to Data Path request with invalid "
                    f"value: {parse_result.error}",
                    data=raw_kv_entry,
                    endpoint=endpoint,
                ) from parse_result.error
            parsed_key, parsed_value, parsed_versionstamp = parse_result.value

            kv_entry = KvEntry(
                key=key,  # We keep the caller's key which may be a custom type
                value=parsed_value,
                versionstamp=VersionStamp(parsed_versionstamp),
            )
            results.append((key, kv_entry))

        if return_unwrapped:
            assert len(results) == 1
            return results[0]
        return tuple(results)

    async def list(
        self,
        *,
        prefix: AnyKvKeyT | None = None,
        start: AnyKvKeyT | None = None,
        end: AnyKvKeyT | None = None,
        **options: Unpack[KvListOptions],
    ) -> AsyncIterator[ListKvEntry[KvKey]]:
        limit = options.get("limit")
        if limit is not None and limit < 0:
            raise ValueError(f"limit cannot be negative: {limit}")

        batch_size = options.get("batch_size")
        batch_size = min(500, (limit or 100) if batch_size is None else batch_size)
        if batch_size < 1:
            raise ValueError(f"batch_size cannot be < 1: {batch_size}")
        reverse = options.get("reverse") or False
        consistency = options.get("consistency") or ConsistencyLevel.STRONG
        cursor = options.get("cursor")

        prefix = None if prefix is None else self._prepare_key(prefix)
        start = None if start is None else self._prepare_key(start)
        end = None if end is None else self._prepare_key(end)

        read_range = read_range_multi(
            prefix=prefix,
            start=start,
            end=end,
            reverse=reverse,
            limit=batch_size,
        )

        context = ListContext(
            prefix=prefix,
            start=start,
            end=end,
            packed_start=read_range.start,
            packed_end=read_range.end,
            limit=limit,
            cursor=cursor,
            reverse=reverse,
            consistency=consistency,
            batch_size=batch_size,
            cursor_format_type=options.get("cursor_format_type")
            or Base64KeySuffixCursorFormat.from_list_context,
        )

        batch_start: KvKeyTuple | None = None
        if cursor is not None:
            cursor_result = context.cursor_format.get_key_for_cursor(cursor)
            if isinstance(cursor_result, Err):
                raise cursor_result.error
            if not (read_range.start <= pack_key(cursor_result.value) < read_range.end):
                raise InvalidCursor(
                    "cursor is not within the the start and end key range",
                    cursor=cursor,
                )
            batch_start = cursor_result.value

        decoder = self.v8_decoder
        if limit == 0:
            return
        count = 0
        while True:
            if batch_start is not None:
                # With a known limit we can reduce the batch size on the final
                # batch to avoid reading results we can't yield
                count_remaining = None if limit is None else limit - count
                required_batch_size = min(batch_size, count_remaining or batch_size)

                # re-calculate the range to start from the cursor position
                if reverse:
                    # start and end of reversed ranges remain in ascending
                    # order, the order results are returned in is reversed.
                    # So the batch start controls end bound in reverse order.
                    read_range = read_range_multi(
                        prefix=prefix,
                        start=start,
                        end=batch_start,
                        reverse=True,
                        limit=required_batch_size,
                    )
                else:
                    read_range = read_range_multi(
                        prefix=prefix,
                        # The batch_start is the key to start after, unlike the
                        # normal start key which is the (inclusive) key to start
                        # from.
                        exclude_start=True,
                        start=batch_start,
                        end=end,
                        reverse=False,
                        limit=required_batch_size,
                    )

            snapshot_read_result = await self._snapshot_read(
                ranges=[read_range], consistency=consistency
            )
            if isinstance(snapshot_read_result, Err):
                raise snapshot_read_result.error
            result, endpoint = snapshot_read_result.value
            if len(result.ranges) != 1:
                raise ProtocolViolation(
                    f"Server responded with {len(result.ranges)} ranges to "
                    f"request for 1 range",
                    data=result,
                    endpoint=endpoint,
                )

            parsed_key: KvKeyTuple | None = None
            (result_range,) = result.ranges
            for raw_kv_entry in result_range.values:
                parse_result = parse_protobuf_kv_entry(
                    raw_kv_entry, v8_decoder=decoder, le64_type=KvU64
                )
                if isinstance(parse_result, Err):
                    raise ProtocolViolation(
                        f"Server responded to Data Path request with invalid "
                        f"value: {parse_result.error}",
                        data=raw_kv_entry,
                        endpoint=endpoint,
                    ) from parse_result.error
                parsed_key, parsed_value, parsed_versionstamp = parse_result.value

                kv_entry = ListKvEntry(
                    key=KvKey.wrap_tuple_keys(parsed_key),
                    value=parsed_value,
                    versionstamp=VersionStamp(parsed_versionstamp),
                    listing=context,
                )
                yield kv_entry

                count += 1
                if limit is not None and count >= limit:
                    return

            # If the read returned less results than the limit, we must have
            # read all the keys that exist within the key's range. Another read
            # would return an empty set.
            if len(result_range.values) < read_range.limit:
                return

            if parsed_key is None:
                assert len(result_range.values) == 0
                return
            batch_start = parsed_key

    async def _snapshot_read(
        self, ranges: Sequence[ReadRange], *, consistency: ConsistencyLevel
    ) -> _KvSnapshotReadResult:
        read = SnapshotRead(ranges=ranges)
        result: SnapshotReadResult
        endpoint: EndpointInfo
        for delay in attempts(self.retry_delays):
            # return error from this?
            cached_meta = await self.metadata_cache.get()
            if isinstance(cached_meta, Err):
                # In the typical case, errors should only propagate from the
                # metadata cache if they're non-retryable, because the
                # metadata-fetching auth function is expected to be doing its
                # own retrying. However it's possible that its retries are
                # exhausted, or its not configured to retry itself, so it seems
                # reasonable to retry retry-able errors.
                if cached_meta.error.retryable:
                    await asyncio.sleep(delay)
                    continue
                raise cached_meta.error
            endpoints = EndpointSelector(meta=cached_meta.value)
            endpoint = endpoints.get_endpoint(consistency)

            result = await datapath.snapshot_read(
                session=self.session,
                meta=cached_meta.value,
                endpoint=endpoint,
                read=read,
            )
            if isinstance(result, Err):
                if result.error.auto_retry is AutoRetry.AFTER_BACKOFF:
                    await asyncio.sleep(delay)
                    continue
                elif result.error.auto_retry is AutoRetry.AFTER_METADATA_EXCHANGE:
                    self.metadata_cache.purge(cached_meta.value)
                    continue
                assert result.error.auto_retry is AutoRetry.NEVER
                return result
            break
        else:
            assert isinstance(result, Err)
            return result
        assert isinstance(result, Ok)
        return Ok((result.value, endpoint))


_KvSnapshotReadResult: TypeAlias = (
    "Result[tuple[SnapshotReadOutput, EndpointInfo], DataPathError]"
)


@dataclass(frozen=True)
class ListContext:
    prefix: AnyKvKey | None
    start: AnyKvKey | None
    end: AnyKvKey | None
    packed_start: bytes
    packed_end: bytes
    limit: int | None
    cursor: str | None
    reverse: bool
    consistency: ConsistencyLevel
    batch_size: int
    cursor_format_type: Callable[[ListContext], AnyCursorFormat] = field(
        repr=False,
        compare=False,
    )
    cursor_format: AnyCursorFormat = field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "cursor_format", self.cursor_format_type(self))


class AnyCursorFormat(Protocol):
    def get_key_for_cursor(self, cursor: str) -> Result[KvKeyTuple, InvalidCursor]: ...

    def get_cursor_for_key(self, key: AnyKvKey) -> Result[str, ValueError]: ...


@dataclass(frozen=True)
class Base64KeySuffixCursorFormat(AnyCursorFormat):
    r"""
    A cursor format that encodes keys as URL-safe base64.

    Packed key bytes that are common to both the start and end key range are
    omitted from the encoded cursor values. This matches the behaviour of Deno's
    built-in KV.list() cursors.

    Examples
    --------
    >>> from base64 import b64decode
    >>> from denokv.datapath import pack_key_range

    >>> start, end = pack_key_range(prefix=('foo',))
    >>> cf = Base64KeySuffixCursorFormat(packed_start=start, packed_end=end)

    The cursor encodes just the part of the key that isn't in both start and
    end.

    >>> cf.get_cursor_for_key(('foo', 'EXAMPLE'))
    Ok('AkVYQU1QTEUA')
    >>> b64decode('AkVYQU1QTEUA')
    b'\x02EXAMPLE\x00'
    """

    packed_start: bytes
    packed_end: bytes
    redundant_prefix_length: int = field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(
            self,
            "redundant_prefix_length",
            _common_prefix_length(self.packed_start, self.packed_end),
        )

    @classmethod
    def from_list_context(cls, list_context: ListContext) -> Self:
        return cls(
            packed_start=list_context.packed_start, packed_end=list_context.packed_end
        )

    def get_key_for_cursor(self, cursor: str) -> Result[KvKeyTuple, InvalidCursor]:
        try:
            significant_packed_key_bytes = urlsafe_b64decode(cursor)
        except Exception as e:
            err = InvalidCursor("cursor is not valid URL-safe base64", cursor=cursor)
            err.__cause__ = e
            return Err(err)

        packed_key = (
            self.packed_start[: self.redundant_prefix_length]
            + significant_packed_key_bytes
        )

        if not (self.packed_start <= packed_key < self.packed_end):
            return Err(
                InvalidCursor(
                    "cursor is not within the the start and end key range",
                    cursor=cursor,
                )
            )

        try:
            unpacked_key = unpack(packed_key)
            if not is_kv_key_tuple(unpacked_key):
                return Err(
                    InvalidCursor("cursor contains invalid part types", cursor=cursor)
                )
            return Ok(unpacked_key)
        except Exception as e:
            err = InvalidCursor(
                "cursor is not a valid suffix for the start and end keys", cursor=cursor
            )
            err.__cause__ = e
            return Err(err)

    def get_cursor_for_key(self, key: AnyKvKey) -> Result[str, ValueError]:
        packed_key = pack_key(key)
        if not (self.packed_start <= packed_key < self.packed_end):
            return Err(ValueError("key is not within the start and end keys"))

        significant_packed_key_bytes = packed_key[self.redundant_prefix_length :]
        return Ok(urlsafe_b64encode(significant_packed_key_bytes).decode())


def _common_prefix_length(a: Sequence[object], b: Sequence[object]) -> int:
    """
    Get the number of elements that are the same from the start of two sequences.

    >>> _common_prefix_length('abc', 'abd')
    2
    >>> _common_prefix_length('abc', 'xyz')
    0
    >>> _common_prefix_length('abc', 'abc')
    3
    >>> _common_prefix_length('', '')
    0
    """
    match_length = 0
    for match_length, (_a, _b) in enumerate(zip(a, b), start=1):
        if _a != _b:
            match_length -= 1
            break
    return match_length


def _is_python_running_in_interactive_environment() -> bool:
    import sys

    # sys.ps1 is only set in interactive environments:
    #   https://stackoverflow.com/a/64523765/693728
    return hasattr(sys, "ps1")


OpenKvFinalize: TypeAlias = Literal[True, False, "interactive"]


async def open_kv(
    target: URL | str | KvCredentials,
    *,
    access_token: str | None = None,
    session: aiohttp.ClientSession | None = None,
    flags: KvFlags | None = None,
    finalize: OpenKvFinalize | None = None,
) -> Kv:
    """
    Create a connection to a KV database.

    [yarl]: `yarl.URL`
    [DEFAULT_KV_FLAGS]: `denokv.kv.DEFAULT_KV_FLAGS`

    Parameters
    ----------
    target
        The Deno KV database server to connect to. Can be a string or [yarl] URL.
    access_token
        The secret access token to authenticate to the target database with.
        Default: The environment variable `DENO_KV_ACCESS_TOKEN` is read.
    session
        The HTTP client session to use to communicate with the database.
        Default: A new session is created.
    flags
        Enable/disable flags that change Kv behaviour. Default: [DEFAULT_KV_FLAGS]
    finalize
        Whether to create a finalizer to automatically close the Kv instance at
        exit, or when out of scope. If set to 'interactive', a finalizer is
        created only if Python is running as an interactive session.
        Default: ['interactive']

    Notes
    -----
    Although this not an async function, it must be run in the context of an
    asyncio event loop when `session` is not provided, because creating a
    aiohttp.ClientSession requires a loop.
    """
    if isinstance(target, str):
        try:
            target = URL(target)
        except ValueError as e:
            raise ValueError(
                f"Cannot open KV database: target argument str is not a "
                f"valid URL: {e}"
            ) from e
    if isinstance(target, URL):
        if access_token is None:
            access_token = environ.get("DENO_KV_ACCESS_TOKEN") or None
        if access_token is None:
            raise ValueError(
                "Cannot open KV database: access_token argument is None and "
                "DENO_KV_ACCESS_TOKEN environment variable is not set"
            )

        target = KvCredentials(server_url=target, access_token=access_token)
    if finalize not in (True, False, "interactive", None):
        raise ValueError("finalize must be True, False, None or 'interactive'")
    finalize = "interactive" if finalize is None else finalize

    session = session or aiohttp.ClientSession()
    retry = ExponentialBackoff()
    auth = Authenticator(session=session, retry_delays=retry, credentials=target)
    kv = Kv(session=session, auth=auth, retry=retry, flags=flags)
    if finalize is True or (
        finalize == "interactive" and _is_python_running_in_interactive_environment()
    ):
        kv.create_finalizer()
    return kv



from __future__ import annotations

import itertools
import random as random_
import time as time_
from dataclasses import dataclass
from enum import IntEnum
from itertools import count
from typing import TYPE_CHECKING
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import Literal

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

Backoff: TypeAlias = Iterable[float]
"""
A backoff strategy for retrying failed operations.

The strategy is simply an iterable of delays (in seconds) to wait for before
retrying after each failure. The strategy may stop iteration after a timeout, or
keep producing delays indefinitely.
"""


@dataclass(init=False)
class ExponentialBackoff(Iterable[float]):
    """
    A `Backoff` strategy that generates randomised, exponentially-increasing delays.

    The delays are $$initial_interval_seconds * multiplier ** attempt$$, with
    the actual delay being randomised from 0 to the calculated delay.

    The `max_interval_seconds` limit the max delay per iteration, and
    `max_elapsed_seconds` limit $$now + delay <= start + max_elapsed_seconds$$.
    Iteration stops once the time reaches $$start + max_elapsed_seconds$$.

    Notes
    -----
    1. The jitter ranging from 0 to the interval baseline delay is based on
       [Exponential Backoff And Jitter].
    2. The default intervals, multipliers and interval/elapsed maximums are those
       used by [google-http-java-client] (but jitter here is different per note
       1.).

    [Exponential Backoff And Jitter]: https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    [google-http-java-client]: https://github.com/googleapis/google-http-java-client/blob/c39e63c/google-http-client/src/main/java/com/google/api/client/util/ExponentialBackOff.java

    Examples
    --------
    >>> import random
    >>> random.seed(43)
    >>> backoff = ExponentialBackoff()
    >>> for i, delay in zip(range(1, 11), backoff):
    ...     print(f'delay {i}: {delay:.2f} seconds')
    delay 1: 0.02 seconds
    delay 2: 0.52 seconds
    delay 3: 0.16 seconds
    delay 4: 0.78 seconds
    delay 5: 1.70 seconds
    delay 6: 3.01 seconds
    delay 7: 2.58 seconds
    delay 8: 4.26 seconds
    delay 9: 0.25 seconds
    delay 10: 8.31 seconds

    Random jitter is not intended to be configurable, but observing behaviour
    without jitter is illustrative:

    >>> i = 1
    >>> now = 0
    >>> backoff = ExponentialBackoff(random=lambda: 1, time=lambda: now,
    ...                              max_elapsed_seconds=20)
    >>> for delay in backoff:
    ...     print(f'delay {i}: {delay:.2f} seconds ({now:.2f} seconds elapsed)')
    ...
    ...     now += delay
    ...     i += 1
    ... else:
    ...     print(f'end — no more delays available ({now:.2f} seconds elapsed)')
    delay 1: 0.50 seconds (0.00 seconds elapsed)
    delay 2: 0.75 seconds (0.50 seconds elapsed)
    delay 3: 1.12 seconds (1.25 seconds elapsed)
    delay 4: 1.69 seconds (2.38 seconds elapsed)
    delay 5: 2.53 seconds (4.06 seconds elapsed)
    delay 6: 3.80 seconds (6.59 seconds elapsed)
    delay 7: 5.70 seconds (10.39 seconds elapsed)
    delay 8: 3.91 seconds (16.09 seconds elapsed)
    end — no more delays available (20.00 seconds elapsed)
    """

    multiplier: float
    initial_interval_seconds: float
    max_interval_seconds: float
    max_elapsed_seconds: float
    random: Callable[[], float]
    time: Callable[[], float]

    def __init__(
        self,
        multiplier: float = 1.5,
        initial_interval_seconds: float = 0.5,
        max_interval_seconds: float = 15 * 60,
        max_elapsed_seconds: float = 15 * 60 * 60,
        random: Callable[[], float] = random_.random,
        time: Callable[[], float] = time_.time,
    ) -> None:
        if multiplier < 1:
            raise ValueError("multiplier must be >= 1")
        if initial_interval_seconds <= 0:
            raise ValueError("initial_interval_seconds must be > 0")
        if max_interval_seconds <= 0:
            raise ValueError("max_interval_seconds must be > 0")
        if max_elapsed_seconds < 0:
            raise ValueError("max_elapsed_seconds must be >= 0")

        self.multiplier = multiplier
        self.initial_interval_seconds = initial_interval_seconds
        self.max_interval_seconds = max_interval_seconds
        self.max_elapsed_seconds = max_elapsed_seconds
        self.random = random
        self.time = time

    def _next_jitter(self) -> float:
        sample = self.random()
        assert 0 <= sample <= 1, "random sample out of range"
        return sample

    def __iter__(self) -> Iterator[float]:
        start = now = self.time()
        limit = start + self.max_elapsed_seconds
        base_interval = 0.0

        for i in count():
            if now >= limit:
                break

            # Stop exponentiating once we hit the max, otherwise we'll overflow
            # if we have lots of iterations with a small max_interval_seconds.
            if base_interval < self.max_interval_seconds:
                base_interval = self.initial_interval_seconds * self.multiplier**i
                base_interval = min(self.max_interval_seconds, base_interval)
            randomised_interval = base_interval * self._next_jitter()  # 0..1
            interval = min(limit - now, randomised_interval)
            yield interval

            now = self.time()


class FinalDelayEnum(IntEnum):
    FinalDelay = 0

    def __repr__(self) -> str:
        return self.name


FinalDelayType: TypeAlias = Literal[FinalDelayEnum.FinalDelay]
FinalDelay: FinalDelayType = FinalDelayEnum.FinalDelay


def attempts(backoff: Iterable[float]) -> Iterable[float | FinalDelayType]:
    """
    Generate the retry-delays in a backoff strategy, ensuring a final attempt.

    The delays from `backoff` will all be generated, followed by a `FinalDelay`
    value, which is also the number 0. Callers can use this to detect the last
    attempt if they care, or just sleep for each delay, with the last always
    taking 0 seconds.

    Because the final delay will always be the 0-second `FinalDelay`, it's fine
    for retry code to sleep for 0 and then have the `for delay in backoff` loop
    finish, using an `else:` clause on the loop to handle exhaustion of retries.
    See the examples section.

    Notes
    -----
    Iterating a `backoff` iterable as `for delay in backoff:` without
    `attempts(backoff)` is a mistake, because the final delay in `backoff` will
    not be followed by another iteration of the loop body.

    FinalDelay: `denokv.backoff.FinalDelay`

    Examples
    --------
    >>> list(attempts([1, 1.5, 3]))
    [1, 1.5, 3, FinalDelay]
    >>> list(attempts([]))
    [FinalDelay]

    >>> for delay in attempts([1.5]):
    ...     print('attempting operation...')
    ...     print(f'sleeping for {delay} before next attempt')
    ... else:
    ...     print('no more attempts available')
    attempting operation...
    sleeping for 1.5 before next attempt
    attempting operation...
    sleeping for 0 before next attempt
    no more attempts available

    >>> for delay in attempts([1.5]):
    ...     print('attempting operation...')
    ...     if delay is not FinalDelay:
    ...         print(f'sleeping for {delay} before next attempt')
    ...     else:
    ...         print('no more attempts available')
    attempting operation...
    sleeping for 1.5 before next attempt
    attempting operation...
    no more attempts available
    """
    return itertools.chain(backoff, (FinalDelay,))
